"""Testing formula 5.21 of EN 1993-5:2007."""

import pytest

from blueprints.codes.eurocode.en_1993_5_2007.chapter_5_ultimate_limit_states.formula_5_21 import Form5Dot21ReducedMomentResistanceClass2UProfiles
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot21ReducedMomentResistanceClass2UProfiles:
    """Validation for formula 5.21 from EN 1993-5:2007."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        m_c_rd = 50.0  # kNm
        n_ed = 20.0  # kN
        n_pl_rd = 40.0  # kN

        # Object to test
        formula = Form5Dot21ReducedMomentResistanceClass2UProfiles(m_c_rd=m_c_rd, n_ed=n_ed, n_pl_rd=n_pl_rd)

        # Expected result, manually calculated
        manually_calculated_result = 33.25  # kNm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("m_c_rd", "n_ed", "n_pl_rd"),
        [
            (-50.0, 20.0, 40.0),  # m_c_rd is negative
            (50.0, -20.0, 40.0),  # n_ed is negative
        ],
    )
    def test_raise_error_when_negative_values_are_given(self, m_c_rd: float, n_ed: float, n_pl_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form5Dot21ReducedMomentResistanceClass2UProfiles(m_c_rd=m_c_rd, n_ed=n_ed, n_pl_rd=n_pl_rd)

    @pytest.mark.parametrize(
        ("m_c_rd", "n_ed", "n_pl_rd"),
        [
            (50.0, 20.0, 0.0),  # n_pl_rd is zero
            (50.0, 20.0, -40.0),  # n_pl_rd is negative
        ],
    )
    def test_raise_error_when_less_or_equal_to_zero_values_are_given(self, m_c_rd: float, n_ed: float, n_pl_rd: float) -> None:
        """Test invalid values."""
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot21ReducedMomentResistanceClass2UProfiles(m_c_rd=m_c_rd, n_ed=n_ed, n_pl_rd=n_pl_rd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"M_{N,Rd} = 1.33 \cdot M_{c,Rd} \cdot \left(1 - \frac{N_{Ed}}{N_{pl,Rd}}\right) = "
                r"1.33 \cdot 50.000 \cdot \left(1 - \frac{20.000}{40.000}\right) = 33.250 \ kNm",
            ),
            ("short", r"M_{N,Rd} = 33.250 \ kNm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        m_c_rd = 50.0  # kNm
        n_ed = 20.0  # kN
        n_pl_rd = 40.0  # kN

        # Object to test
        latex = Form5Dot21ReducedMomentResistanceClass2UProfiles(m_c_rd=m_c_rd, n_ed=n_ed, n_pl_rd=n_pl_rd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 3.9 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS


class Form3Dot9DryingShrinkage(Formula):
    """Class representing formula 3.9 for the calculation of the drying shrinkage."""

    label = "3.9"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        beta_ds_tt_s: DIMENSIONLESS,
        k_h: DIMENSIONLESS,
        epsilon_cd_0: DIMENSIONLESS,
    ) -> None:
        r"""[$\epsilon_{cd}(t)$] Development of the drying shrinkage [$-$].

        EN 1992-1-1:2004 art.3.1.4(6) - Formula (3.9)

        Parameters
        ----------
        beta_ds_tt_s : DIMENSIONLESS
            [$\beta_{ds}(t, t_s)$] Coefficient that depends on the age t (in days) of the concrete for the drying shrinkage [$-$].
        k_h : DIMENSIONLESS
            [$k_h$] Coefficient depending on the fictional thickness $h_0$ following table 3.3 [$-$].
            $h_0 = 100 \rightarrow k_h = 1.0$
            $h_0 = 200 \rightarrow k_h = 0.85$
            $h_0 = 300 \rightarrow k_h = 0.75$
            $h_0 \geq 500 \rightarrow k_h = 0.70$
        epsilon_cd_0 : DIMENSIONLESS
            [$\epsilon_{cd,0}$] Nominal unobstructed drying shrinkage, formula in appendix B or use table 3.2 [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.beta_ds_tt_s = beta_ds_tt_s
        self.k_h = k_h
        self.epsilon_cd_0 = epsilon_cd_0

    @staticmethod
    def _evaluate(
        beta_ds_tt_s: DIMENSIONLESS,
        k_h: DIMENSIONLESS,
        epsilon_cd_0: DIMENSIONLESS,
    ) -> float:
        """Evaluates the formula, for more information see the __init__ method."""
        if beta_ds_tt_s < 0:
            raise ValueError(f"Negative beta_ds_tt_s: {beta_ds_tt_s}. beta_ds_tt_s cannot be negative")
        if k_h < 0:
            raise ValueError(f"Negative k_h: {k_h}. k_h cannot be negative")
        return beta_ds_tt_s * k_h * epsilon_cd_0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.9."""
        return LatexFormula(
            return_symbol=r"\epsilon_{cd}(t)",
            result=f"{self:.{n}f}",
            equation=r"\beta_{ds}(t,t_s) \cdot k_h \cdot \epsilon_{cd,0}",
            numeric_equation=rf"{self.beta_ds_tt_s:.{n}f} \cdot {self.k_h:.{n}f} \cdot {self.epsilon_cd_0:.{n}f}",
            comparison_operator_label="=",
        )



"""Testing formula 5.25 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_25 import Form5Dot25AxialForceCorrectionFactor
from blueprints.validations import NegativeValueError


class TestForm5Dot25AxialForceCorrectionFactor:
    """Validation for formula 5.25 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        n = 0.5  # -

        # Object to test
        form_5_25 = Form5Dot25AxialForceCorrectionFactor(n=n)

        # Expected result, manually calculated
        manually_calculated_result = 0.15  # -

        assert form_5_25 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_n_is_given(self) -> None:
        """Test a negative value for n."""
        # Example values
        n = -0.5

        with pytest.raises(NegativeValueError):
            Form5Dot25AxialForceCorrectionFactor(n=n)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            ("complete", r"k_{2} = \min(0.20; n \cdot 0.30) = \min(0.20; 0.500 \cdot 0.30) = 0.150"),
            ("short", "k_{2} = 0.150"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n = 0.5  # -

        # Object to test
        form_5_25_latex = Form5Dot25AxialForceCorrectionFactor(n=n).latex()

        actual = {
            "complete": form_5_25_latex.complete,
            "short": form_5_25_latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Formula 7.7n from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import MM, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form7Dot7nMaxBarDiameterTension(Formula):
    r"""Class representing formula 7.7n for the calculation of [$⌀_s$]."""

    label = "7.7n"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        diam_s_star: MM,
        f_ct_eff: MPA,
        h_cr: MM,
        h: MM,
        d: MM,
    ) -> None:
        r"""[$⌀_s$] Calculation of the maximum bar diameter for tension [$mm$].

        EN 1992-1-1:2004 art.7.3.3(2) - Formula (7.7n)

        Parameters
        ----------
        diam_s_star : MM
            [$⌀^*_s$] Maximum bar size given in the Table 7.2N [$mm$].
        f_ct_eff : MPA
            [$f_{ct,eff}$] Mean value of the tensile strength of the concrete effective at the time
            when the cracks may first be expected to occur [$MPa$].
        h_cr : MM
            [$h_{cr}$] Depth of the tensile zone immediately prior to cracking, considering the characteristic values
            of prestress and axial forces under the quasi-permanent combination of actions [$mm$].
        h : MM
            [$h$] Overall depth of the section [$mm$].
        d : MM
            [$d$] Effective depth to the centroid of the outer layer of reinforcement [$mm$].
        """
        super().__init__()
        self.diam_s_star = diam_s_star
        self.f_ct_eff = f_ct_eff
        self.h_cr = h_cr
        self.h = h
        self.d = d

    @staticmethod
    def _evaluate(
        diam_s_star: MM,
        f_ct_eff: MPA,
        h_cr: MM,
        h: MM,
        d: MM,
    ) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(diam_s_star=diam_s_star, f_ct_eff=f_ct_eff, h_cr=h_cr, h=h, d=d)
        denominator: MM = 2 * (h - d)
        raise_if_less_or_equal_to_zero(denominator=denominator)

        return diam_s_star * (f_ct_eff / 2.9) * (h_cr) / (8 * (h - d))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.7n."""
        _equation: str = r"⌀^*_s \cdot \left(\frac{f_{ct,eff}}{2.9}\right) \cdot \left(\frac{h_{cr}}{8 \cdot ( h - d)}\right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"⌀^*_s": f"{self.diam_s_star:.{n}f}",
                r"f_{ct,eff}": f"{self.f_ct_eff:.{n}f}",
                r"h_{cr}": f"{self.h_cr:.{n}f}",
                r" h": f" {self.h:.{n}f}",
                r" d": f" {self.d:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"⌀_s",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm",
        )



"""Formula 5.13 from EN 1993-5:2007 Chapter 5 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1993_5_2007 import EN_1993_5_2007
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, KN, KNM, MM2, MPA
from blueprints.unit_conversion import N_TO_KN
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot13SimplifiedBucklingCheck(Formula):
    """Class representing formula 5.13 for combined axial force and bending moment check. Only valid for class 1, 2 and 3."""

    label = "5.13"
    source_document = EN_1993_5_2007

    def __init__(
        self,
        n_ed: KN,
        m_ed: KNM,
        a: MM2,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
        gamma_m1: DIMENSIONLESS,
        chi: DIMENSIONLESS,
        m_c_rd: KNM,
    ) -> None:
        r"""
        Simplified buckling check for class 1, 2 and 3.

        Provided that the boundary conditions are supplied by elements (anchor, earth support, capping beam, etc.)
        that give positional restraint corresponding to the non-sway buckling mode, this check may be used:

        Parameters
        ----------
        n_ed : KN
            [$kN$] Design axial force [$kN$].
        m_ed : KNM
            [$kNm$] Design bending moment [$kNm$].
        a : MM2
            [$mm^2$] Cross-sectional area [$mm^2$].
        f_y : MPA
            [$MPa$] Yield strength of the material [$MPa$].
        gamma_m0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial factor according to 5.1.1 (4) [$-$].
        gamma_m1 : DIMENSIONLESS
            [$\gamma_{M1}$] Partial factor according to 5.1.1 (4) [$-$].
        chi : DIMENSIONLESS
            [$\chi$] Buckling coefficient from 6.3.1.2 of EN 1993-1-1 [$-$].
        m_c_rd : KNM
            [$kNm$] Design moment resistance of the cross-section [$kNm$]. See 5.2.2 (2).
            This can also be calculated using Form5Dot2DesignMomentResistanceClass1Or2 or Form5Dot3DesignMomentResistanceClass3.
        """
        super().__init__()
        self.n_ed = n_ed
        self.m_ed = m_ed
        self.a = a
        self.f_y = f_y
        self.gamma_m0 = gamma_m0
        self.gamma_m1 = gamma_m1
        self.chi = chi
        self.m_c_rd = m_c_rd

    @staticmethod
    def _evaluate(
        n_ed: KN,
        m_ed: KNM,
        a: MM2,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
        gamma_m1: DIMENSIONLESS,
        chi: DIMENSIONLESS,
        m_c_rd: KNM,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            n_ed=n_ed,
            m_ed=m_ed,
        )
        raise_if_less_or_equal_to_zero(
            a=a,
            f_y=f_y,
            chi=chi,
            m_c_rd=m_c_rd,
            gamma_m0=gamma_m0,
            gamma_m1=gamma_m1,
        )
        n_pl_rd: KN = a * f_y / gamma_m0 * N_TO_KN
        return (n_ed / (chi * n_pl_rd * (gamma_m0 / gamma_m1)) + 1.15 * m_ed / (m_c_rd * (gamma_m0 / gamma_m1))) <= 1.0

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 5.13."""
        _equation: str = (
            r"\frac{N_{Ed}}{\chi \cdot (A \cdot f_{y} / \gamma_{M0}) \cdot \left( \frac{\gamma_{M0}}{\gamma_{M1}} \right)} + "
            r"1.15 \cdot \frac{M_{Ed}}{M_{c,Rd} \cdot \left( \frac{\gamma_{M0}}{\gamma_{M1}} \right)} \leq 1.0"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "N_{Ed}": f"{self.n_ed:.{n}f}",
                "M_{Ed}": f"{self.m_ed:.{n}f}",
                "A": f"{self.a:.{n}f} / 10^6",
                "f_{y}": f"{self.f_y:.1f} \\cdot 10^3",
                "M_{c,Rd}": f"{self.m_c_rd:.{n}f}",
                r"\gamma_{M0}": f"{self.gamma_m0:.1f}",
                r"\gamma_{M1}": f"{self.gamma_m1:.1f}",
                r"\chi": f"{self.chi:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing sub-formula 1 for 8.8N of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_8n import (
    SubForm8Dot8nConcreteStress,
    SubForm8Dot8nDesignLengthOfTransverseBar,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestSubForm8Dot8nDesignLengthOfTransverseBar:
    """Validation for sub-formula 8.8N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        diameter_t = 16  # mm
        f_yd = 500  # MPa
        sigma_td = 60  # MPa
        l_t = 100  # mm
        sub_form_8_8n_1 = SubForm8Dot8nDesignLengthOfTransverseBar(
            diameter_t=diameter_t,
            f_yd=f_yd,
            sigma_td=sigma_td,
            l_t=l_t,
        )
        # Expected result, manually calculated
        manually_result = 53.578104
        assert sub_form_8_8n_1 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_evaluation_upper_limit(self) -> None:
        """Test the evaluation of the result if the upper limit is reached."""
        diameter_t = 32  # mm
        f_yd = 500  # MPa
        sigma_td = 60  # MPa
        l_t = 50  # mm
        sub_form_8_8n_1 = SubForm8Dot8nDesignLengthOfTransverseBar(
            diameter_t=diameter_t,
            f_yd=f_yd,
            sigma_td=sigma_td,
            l_t=l_t,
        )
        # Expected result, manually calculated
        manually_result = 50
        assert sub_form_8_8n_1 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_raise_error_when_diameter_t_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when diameter_t is negative."""
        # Example values
        diameter_t = -16  # mm
        f_yd = 500  # MPa
        sigma_td = 60  # MPa
        l_t = 100  # mm

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nDesignLengthOfTransverseBar(
                diameter_t=diameter_t,
                f_yd=f_yd,
                sigma_td=sigma_td,
                l_t=l_t,
            )

    def test_raise_error_when_f_yd_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when f_yd is negative."""
        # Example values
        diameter_t = 16  # mm
        f_yd = -500  # MPa
        sigma_td = 60  # MPa
        l_t = 100  # mm

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nDesignLengthOfTransverseBar(
                diameter_t=diameter_t,
                f_yd=f_yd,
                sigma_td=sigma_td,
                l_t=l_t,
            )

    def test_raise_error_when_sigma_td_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when sigma_td is negative."""
        # Example values
        diameter_t = 16  # mm
        f_yd = 500  # MPa
        sigma_td = -60  # MPa
        l_t = 100  # mm

        with pytest.raises(LessOrEqualToZeroError):
            SubForm8Dot8nDesignLengthOfTransverseBar(
                diameter_t=diameter_t,
                f_yd=f_yd,
                sigma_td=sigma_td,
                l_t=l_t,
            )

    def test_raise_error_when_sigma_td_is_zero(self) -> None:
        """Test if a NegativeValueError is raised when sigma_td is zero."""
        # Example values
        diameter_t = 16  # mm
        f_yd = 500  # MPa
        sigma_td = 0  # MPa
        l_t = 100  # mm

        with pytest.raises(LessOrEqualToZeroError):
            SubForm8Dot8nDesignLengthOfTransverseBar(
                diameter_t=diameter_t,
                f_yd=f_yd,
                sigma_td=sigma_td,
                l_t=l_t,
            )

    def test_raise_error_when_l_t_is_negative(self) -> None:
        """Test if a NegativeValueError is raised when l_t is negative."""
        # Example values
        diameter_t = 16  # mm
        f_yd = 500  # MPa
        sigma_td = 60  # MPa
        l_t = -100  # mm

        with pytest.raises(NegativeValueError):
            SubForm8Dot8nDesignLengthOfTransverseBar(
                diameter_t=diameter_t,
                f_yd=f_yd,
                sigma_td=sigma_td,
                l_t=l_t,
            )

    def test_integration_with_sub_form_8_8n_concrete_stress(self) -> None:
        """Test the integration with sub-formula 8.8 for calculating concrete stress."""
        # Example values
        diameter_t = 16  # mm
        f_yd = 500  # MPa
        l_t = 100  # mm
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0.5  # -
        f_cd = 25  # MPa
        sigma_td = SubForm8Dot8nConcreteStress(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            y_function=y_function,
            f_cd=f_cd,
        )

        # Object to test
        form_8_8n = SubForm8Dot8nDesignLengthOfTransverseBar(
            diameter_t=diameter_t,
            f_yd=f_yd,
            sigma_td=sigma_td,
            l_t=l_t,
        )

        manually_calculated_result = 69.950631942

        assert form_8_8n == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                (
                    r"l_{td} = \min\left(l_t, 1.16 \cdot Ø_t \cdot ({\frac{f_{yd}}{\sigma_{td}}})^{0.5} \right) = "
                    r"\min\left(100.00, 1.16 \cdot 16.00 \cdot ({\frac{500.00}{35.20}})^{0.5} \right) = 69.95"
                ),
            ),
            ("short", r"l_{td} = 69.95"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        diameter_t = 16  # mm
        f_yd = 500  # MPa
        l_t = 100  # mm
        f_ctd = 2.6  # MPa
        sigma_cm = 15  # MPa
        y_function = 0.5  # -
        f_cd = 25  # MPa
        sigma_td = SubForm8Dot8nConcreteStress(
            f_ctd=f_ctd,
            sigma_cm=sigma_cm,
            y_function=y_function,
            f_cd=f_cd,
        )

        # Object to test
        sub_form_8_8n_1_latex = SubForm8Dot8nDesignLengthOfTransverseBar(
            diameter_t=diameter_t,
            f_yd=f_yd,
            sigma_td=sigma_td,
            l_t=l_t,
        ).latex()

        actual = {"complete": sub_form_8_8n_1_latex.complete, "short": sub_form_8_8n_1_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 6.47 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM, MM2, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class SubForm6Dot47FactorK(Formula):
    r"""Class representing the sub-formula which calculates the factor [$k$] for formula 6.47 ."""

    label = "6.47 (factor k)"
    source_document = EN_1992_1_1_2004

    def __init__(self, d: MM) -> None:
        r"""[$k$] Calculation of factor k.

        EN 1992-1-1:2004 art.6.4.4(1) - Factor k for Formula (6.47)

        Parameters
        ----------
        d : MM
            [$d$] Effective depth [$mm$].
        """
        super().__init__()
        self.d = d

    @staticmethod
    def _evaluate(d: MM) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(d=d)
        return min(1 + np.sqrt(200 / d), 2.0)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for sub-formula 6.47 (factor k)."""
        _equation: str = r"\min \left( 1 + \sqrt{\frac{200}{d}}, 2.0 \right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"d": f"{self.d:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"k",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )


class SubForm6Dot47FactorRhoL(Formula):
    r"""Class representing the sub-formula which calculates the factor [$\rho_l$] for formula 6.47 ."""

    label = "6.47 (factor rho_l)"
    source_document = EN_1992_1_1_2004

    def __init__(self, rho_ly: DIMENSIONLESS, rho_lz: DIMENSIONLESS) -> None:
        r"""[$\rho_l$] Calculation of factor [$\rho_l$].

        EN 1992-1-1:2004 art.6.4.4(1) - Factor rho_l for Formula (6.47)

        Parameters
        ----------
        rho_ly : DIMENSIONLESS
            [$\rho_{ly}$] Related to the bonded tension steel in y- drection. The value $\rho_ly$ should be calculated as mean values taking
            into account a slab width equal to the column width plus 3d each side [$-$].
        rho_lz : DIMENSIONLESS
            [$\rho_{lz}$] Related to the bonded tension steel in z- drection. The value $\rho_lz$ should be calculated as mean values taking
            into account a slab width equal to the column width plus 3d each side [$-$].
        """
        super().__init__()
        self.rho_ly = rho_ly
        self.rho_lz = rho_lz

    @staticmethod
    def _evaluate(rho_ly: DIMENSIONLESS, rho_lz: DIMENSIONLESS) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(rho_ly=rho_ly, rho_lz=rho_lz)
        return min(np.sqrt(rho_ly * rho_lz), 0.02)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for sub-formula 6.47 (factor rho_l)."""
        _equation: str = r"\min \left( \sqrt{\rho_{ly} \cdot \rho_{lz}}, 0.02 \right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\rho_{ly}": f"{self.rho_ly:.{n}f}",
                r"\rho_{lz}": f"{self.rho_lz:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\rho_l",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )


class SubForm6Dot47FactorSigmaCp(Formula):
    r"""Class representing the sub-formula which calculates the factor [$\sigma_{cp}$] for formula 6.47,."""

    label = "6.47 (factor sigma_cp)"
    source_document = EN_1992_1_1_2004

    def __init__(self, sigma_cy: MPA, sigma_cz: MPA) -> None:
        r"""[$\sigma_{cp}$] Calculation of factor [$\sigma_{cp}$].

        EN 1992-1-1:2004 art.6.4.4(1) - Factor sigma_cp for Formula (6.47)

        Parameters
        ----------
        sigma_cy : MPA
            [$\sigma_{cy}$] Normal concrete stress in the critical section in the y-direction [$MPa$], Positive if compression.
            See equation SubForm6Dot47FactorSigmaCy.
        sigma_cz : MPA
            [$\sigma_{cz}$] Normal concrete stress inm the critical section in the z-direction [$MPa$], Positive if compression.
            See equation SubForm6Dot47FactorSigmaCz.
        """
        super().__init__()
        self.sigma_cy = sigma_cy
        self.sigma_cz = sigma_cz

    @staticmethod
    def _evaluate(sigma_cy: MPA, sigma_cz: MPA) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(sigma_cy=sigma_cy, sigma_cz=sigma_cz)
        return (sigma_cy + sigma_cz) / 2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for sub-formula 6.47 (sigma_cp)."""
        _equation: str = r"\frac{\sigma_{cy} + \sigma_{cz}}{2}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\sigma_{cy}": f"{self.sigma_cy:.{n}f}",
                r"\sigma_{cz}": f"{self.sigma_cz:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\sigma_{cp}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )


class SubForm6Dot47FactorSigmaCy(Formula):
    r"""Class representing the sub-formula which calculates the factor [$\sigma_{cy}$] for formula 6.47."""

    label = "6.47 (factor sigma_cy)"
    source_document = EN_1992_1_1_2004

    def __init__(self, n_ed_y: N, a_cy: MM2) -> None:
        r"""[$\sigma_{cy}$] Calculation of factor [$\sigma_{cy}$].

        EN 1992-1-1:2004 art.6.4.4(1) - Factor sigma_cy for Formula (6.47)

        Parameters
        ----------
        n_ed_y : N
            [$N_{Ed,y}$] Longitudinal forces across the full bay for internal columns and the logintudinal force across
            the control section for edge columns. The force may be from a load or prestressing action [$N$].
        a_cy : MM2
            [$A_{cy}$] Cross-sectional area in y-direction [$mm^2$].
        """
        super().__init__()
        self.n_ed_y = n_ed_y
        self.a_cy = a_cy

    @staticmethod
    def _evaluate(n_ed_y: N, a_cy: MM2) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(a_cy=a_cy)
        raise_if_negative(n_ed_y=n_ed_y)
        return n_ed_y / a_cy

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for sub-formula 6.47 (sigma_cy)."""
        _equation: str = r"\frac{N_{Ed,y}}{A_{cy}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed,y}": f"{self.n_ed_y:.{n}f}",
                r"A_{cy}": f"{self.a_cy:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\sigma_{cy}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )


class SubForm6Dot47FactorSigmaCz(Formula):
    r"""Class representing the sub-formula which calculates the factor [$\sigma_{cz}$] for formula 6.47."""

    label = "6.47 (factor sigma_cz)"
    source_document = EN_1992_1_1_2004

    def __init__(self, n_ed_z: N, a_cz: MM2) -> None:
        r"""[$\sigma_{cz}$] Calculation of factor [$\sigma_{cz}$].

        EN 1992-1-1:2004 art.6.4.4(1) - Factor sigma_cz for Formula (6.47)

        Parameters
        ----------
        n_ed_z : N
            [$N_{Ed,z}$] Longitudinal forces across the full bay for internal columns and the logintudinal force across
            the control section for edge columns. The force may be from a load or prestressing action [$N$].
        a_cz : MM2
            [$A_{cz}$] Cross-sectional area in z-direction [$mm^2$].
        """
        super().__init__()
        self.n_ed_z = n_ed_z
        self.a_cz = a_cz

    @staticmethod
    def _evaluate(n_ed_z: N, a_cz: MM2) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(a_cz=a_cz)
        raise_if_negative(n_ed_z=n_ed_z)
        return n_ed_z / a_cz

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for sub-formula 6.47 sigma_cz."""
        _equation: str = r"\frac{N_{Ed,z}}{A_{cz}}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"N_{Ed,z}": f"{self.n_ed_z:.{n}f}",
                r"A_{cz}": f"{self.a_cz:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\sigma_{cz}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="MPa",
        )



"""Testing formula 5.5 of EN 1993-5:2007."""

import pytest

from blueprints.codes.eurocode.en_1993_5_2007.chapter_5_ultimate_limit_states.formula_5_5 import Form5Dot5PlasticShearResistance
from blueprints.validations import LessOrEqualToZeroError


class TestForm5Dot5PlasticShearResistance:
    """Validation for formula 5.5 from EN 1993-5:2007."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        a_v = 200  # MM2
        f_y = 100  # MPA
        gamma_m_0 = 0.5  # DIMENSIONLESS
        form = Form5Dot5PlasticShearResistance(a_v=a_v, f_y=f_y, gamma_m_0=gamma_m_0)

        # Expected result, manually calculated
        expected = 23.0940  # KN

        assert form == pytest.approx(expected)

    @pytest.mark.parametrize(
        ("a_v", "f_y", "gamma_m_0"),
        [
            (-0.5, 100, 0.5),  # a_v is negative
            (0, 100, 0.5),  # a_v is zero
            (200, -100, 0.5),  # f_y is negative
            (200, 0, 0.5),  # f_y is zero
            (200, 100, -0.5),  # gamma_m_0 is negative
            (200, 100, 0),  # gamma_m_0 is zero
        ],
    )
    def test_raise_error_when_negative_or_zero_n_t_rd_is_given(self, a_v: float, f_y: float, gamma_m_0: float) -> None:
        """Test a negative and zero value for parameters a_v, f_y and gamma_m_0."""
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot5PlasticShearResistance(a_v=a_v, f_y=f_y, gamma_m_0=gamma_m_0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{pl,Rd} = \frac{A_v f_y}{\sqrt{3} \gamma_{M0}} = \frac{200 \cdot 100}{\sqrt{3} \cdot 0.5} = 23.09",
            ),
            ("short", r"V_{pl,Rd} = 23.09"),
            (
                "string",
                r"V_{pl,Rd} = \frac{A_v f_y}{\sqrt{3} \gamma_{M0}} = \frac{200 \cdot 100}{\sqrt{3} \cdot 0.5} = 23.09",
            ),
        ],
    )
    def test_latex_output(self, representation: str, expected: str) -> None:
        """Test the latex implementation."""
        a_v = 200  # MM2
        f_y = 100  # MPA
        gamma_m_0 = 0.5  # DIMENSIONLESS

        form = Form5Dot5PlasticShearResistance(
            a_v=a_v,
            f_y=f_y,
            gamma_m_0=gamma_m_0,
        ).latex()

        actual = {
            "complete": form.complete,
            "short": form.short,
            "string": str(form),
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""IPE Steel Profiles."""

from enum import Enum

from blueprints.type_alias import MM


class IPE(Enum):
    """Geometrical representation of IPE steel profiles."""

    IPE80 = ("IPE80", 80, 46, 3.8, 5.2, 5)
    IPE100 = ("IPE100", 100, 55, 4.1, 5.7, 7)
    IPE120 = ("IPE120", 120, 64, 4.4, 6.3, 7)
    IPE140 = ("IPE140", 140, 73, 4.7, 6.9, 7)
    IPE160 = ("IPE160", 160, 82, 5.0, 7.4, 9)
    IPE180 = ("IPE180", 180, 91, 5.3, 8.0, 9)
    IPE200 = ("IPE200", 200, 100, 5.6, 8.5, 12)
    IPE220 = ("IPE220", 220, 110, 5.9, 9.2, 12)
    IPE240 = ("IPE240", 240, 120, 6.2, 9.8, 15)
    IPE270 = ("IPE270", 270, 135, 6.6, 10.2, 15)
    IPE300 = ("IPE300", 300, 150, 7.1, 10.7, 15)
    IPE330 = ("IPE330", 330, 160, 7.5, 11.5, 18)
    IPE360 = ("IPE360", 360, 170, 8.0, 12.7, 18)
    IPE400 = ("IPE400", 400, 180, 8.6, 13.5, 21)
    IPE450 = ("IPE450", 450, 190, 9.4, 14.6, 21)
    IPE500 = ("IPE500", 500, 200, 10.2, 16.0, 21)
    IPE550 = ("IPE550", 550, 210, 11.1, 17.2, 24)
    IPE600 = ("IPE600", 600, 220, 12.0, 19.0, 24)

    def __init__(self, alias: str, h: MM, b: MM, t_w: MM, t_f: MM, radius: MM) -> None:
        """Initialize IPE profile.

        This method sets the profile's alias, dimensions, and radii.

        Parameters
        ----------
        alias: str
            Profile alias.
        h: MM
            Total height of the profile.
        b: MM
            Total width of the profile.
        t_w: MM
            Web thickness of the profile.
        t_f: MM
            Flange thickness of the profile.
        radius: MM
            Radius of the profile.
        """
        self.alias = alias
        self.top_flange_width = b
        self.top_flange_thickness = t_f
        self.bottom_flange_width = b
        self.bottom_flange_thickness = t_f
        self.total_height = h
        self.web_thickness = t_w
        self.top_radius = radius
        self.bottom_radius = radius



"""Formula 3.8 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS


class Form3Dot8TotalShrinkage(Formula):
    """Class representing formula 3.8 for the calculation of the total shrinkage."""

    label = "3.8"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        epsilon_cd: DIMENSIONLESS,
        epsilon_ca: DIMENSIONLESS,
    ) -> None:
        r"""[$\epsilon_{cs}$] The total shrinkage [-].

        EN 1992-1-1:2004 art.3.1.4(6) - Formula (3.8)

        Parameters
        ----------
        epsilon_cd : DIMENSIONLESS
            [$\epsilon_{cd}$] Drying shrinkage [$-$].
        epsilon_ca : DIMENSIONLESS
            [$\epsilon_{ca}$] Autogene shrinkage [$-$].

        Returns
        -------
        None
        """
        super().__init__()
        self.epsilon_cd = epsilon_cd
        self.epsilon_ca = epsilon_ca

    @staticmethod
    def _evaluate(
        epsilon_cd: DIMENSIONLESS,
        epsilon_ca: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        return epsilon_cd + epsilon_ca

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.8."""
        return LatexFormula(
            return_symbol=r"\epsilon_{cs}",
            result=f"{self:.{n}f}",
            equation=r"\epsilon_{cd} + \epsilon_{ca}",
            numeric_equation=rf"{self.epsilon_cd:.{n}f} + {self.epsilon_ca:.{n}f}",
            comparison_operator_label="=",
        )



"""Testing formula 5.12 of EN 1993-5:2007."""

import pytest

from blueprints.codes.eurocode.en_1993_5_2007.chapter_5_ultimate_limit_states.formula_5_12 import Form5Dot12ElasticCriticalLoad
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot12ElasticCriticalLoad:
    """Validation for formula 5.12 from EN 1993-5:2007."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        e = 210000  # MPA
        i = 5000  # MM4
        beta_d = 1.0  # DIMENSIONLESS
        l = 1000  # MM  # noqa: E741

        # Object to test
        formula = Form5Dot12ElasticCriticalLoad(e=e, i=i, beta_d=beta_d, l=l)

        # Expected result, manually calculated
        manually_calculated_result = 10363.1  # N

        assert formula == pytest.approx(manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("e", "i", "beta_d", "l"),
        [
            (-210000, 5000, 1.0, 1000),  # e is negative
            (210000, -5000, 1.0, 1000),  # i is negative
            (210000, 5000, -1.0, 1000),  # beta_d is negative
            (210000, 5000, 1.0, -1000),  # l is negative
            (210000, 5000, 1.0, 0),  # l is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, e: float, i: float, beta_d: float, l: float) -> None:  # noqa: E741
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot12ElasticCriticalLoad(e=e, i=i, beta_d=beta_d, l=l)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"N_{cr} = \frac{E \cdot I \cdot \beta_D \cdot \pi^2}{l^2} = "
                r"\frac{210000.000 \cdot 5000.000 \cdot 1.000 \cdot \pi^2}{1000.000^2} = 10363.085 \ N",
            ),
            ("short", r"N_{cr} = 10363.085 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e = 210000  # MPA
        i = 5000  # MM4
        beta_d = 1.0  # DIMENSIONLESS
        l = 1000  # MM  # noqa: E741

        # Object to test
        latex = Form5Dot12ElasticCriticalLoad(e=e, i=i, beta_d=beta_d, l=l).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""formula 6.41 from EN 1993-1-1:2005: chapter 6 - ultimate limit state."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, NMM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot41BiaxialBendingCheck(Formula):
    r"""class representing formula 6.41 for bi-axial bending."""

    label = "6.41"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        my_ed: NMM,
        m_n_y_rd: NMM,
        mz_ed: NMM,
        m_n_z_rd: NMM,
        alpha: DIMENSIONLESS,
        beta: DIMENSIONLESS,
    ) -> None:
        r"""Check the bi-axial bending criterion.

        EN 1993-1-1:2005 art.6.2.9.1(6) - formula (6.41)

        Parameters
        ----------
        my_ed : NMM
            [$m_{y,ed}$] design bending moment about the y-axis [$Nmm$].
        m_n_y_rd : NMM
            [$m_{n,y,rd}$] design resistance moment about the y-axis [$Nmm$].
        mz_ed : NMM
            [$m_{z,ed}$] design bending moment about the z-axis [$Nmm$].
        m_n_z_rd : NMM
            [$m_{n,z,rd}$] design resistance moment about the z-axis [$Nmm$].
        alpha : DIMENSIONLESS
            [$\alpha$] exponent for the y-axis term.
        beta : DIMENSIONLESS
            [$\beta$] exponent for the z-axis term.

        In which $\alpha$ and $\beta$ are constants, which may conservatively be taken as unity, otherwise as follows:
        - I and H sections:
            $\alpha = 2$ and $\beta = 5 \cdot n$, but $\beta \geq 1$.
        - Circular hollow sections:
            $\alpha = 2$ and $\beta = 2$.
        - Rectangular hollow sections:
            $\alpha = \beta = \frac{1.66}{1-1.13 \cdot n^2}$, but $\alpha = \beta \leq 6$.

        Where $n = N_{Ed} / N_{pl,Rd}$, see equation 6.38n.
        """
        super().__init__()
        self.my_ed = my_ed
        self.m_n_y_rd = m_n_y_rd
        self.mz_ed = mz_ed
        self.m_n_z_rd = m_n_z_rd
        self.alpha = alpha
        self.beta = beta

    @staticmethod
    def _evaluate(
        my_ed: NMM,
        m_n_y_rd: NMM,
        mz_ed: NMM,
        m_n_z_rd: NMM,
        alpha: DIMENSIONLESS,
        beta: DIMENSIONLESS,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(m_n_y_rd=m_n_y_rd, m_n_z_rd=m_n_z_rd)
        raise_if_negative(my_ed=my_ed, mz_ed=mz_ed, alpha=alpha, beta=beta)

        return (my_ed / m_n_y_rd) ** alpha + (mz_ed / m_n_z_rd) ** beta <= 1

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.41."""
        _equation: str = r"\left[ \frac{M_{y,Ed}}{M_{N,y,Rd}} \right]^{\alpha} + \left[ \frac{M_{z,Ed}}{M_{N,z,Rd}} \right]^{\beta} \leq 1"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                "M_{y,Ed}": f"{self.my_ed:.{n}f}",
                "M_{N,y,Rd}": f"{self.m_n_y_rd:.{n}f}",
                "M_{z,Ed}": f"{self.mz_ed:.{n}f}",
                "M_{N,z,Rd}": f"{self.m_n_z_rd:.{n}f}",
                r"\alpha": f"{self.alpha:.{n}f}",
                r"\beta": f"{self.beta:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                "M_{y,Ed}": rf"{self.my_ed:.{n}f} \ Nmm",
                "M_{N,y,Rd}": rf"{self.m_n_y_rd:.{n}f} \ Nmm",
                "M_{z,Ed}": rf"{self.mz_ed:.{n}f} \ Nmm",
                "M_{N,z,Rd}": rf"{self.m_n_z_rd:.{n}f} \ Nmm",
                r"\alpha": rf"{self.alpha:.{n}f}",
                r"\beta": rf"{self.beta:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula 7.17 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_17 import (
    Form7Dot1MultiplicationFactorLimitSlenderness,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot1MultiplicationFactorLimitSlenderness:
    """Validation for formula 7.17 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        f_yk = 500.0
        a_s_req = 200.0
        a_s_prov = 250.0

        # Object to test
        formula = Form7Dot1MultiplicationFactorLimitSlenderness(f_yk=f_yk, a_s_req=a_s_req, a_s_prov=a_s_prov)

        # Expected result, manually calculated
        manually_calculated_result = 1.25  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("f_yk", "a_s_req", "a_s_prov"),
        [
            (-500.0, 200.0, 250.0),  # f_yk is negative
            (500.0, -200.0, 250.0),  # a_s_req is negative
            (500.0, 200.0, -250.0),  # a_s_prov is negative
            (500.0, 0.0, 250.0),  # a_s_req is zero
            (500.0, 200.0, 0.0),  # a_s_prov is zero
            (0.0, 200.0, 250.0),  # f_yk is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, f_yk: float, a_s_req: float, a_s_prov: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot1MultiplicationFactorLimitSlenderness(f_yk=f_yk, a_s_req=a_s_req, a_s_prov=a_s_prov)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\frac{310}{\sigma_s} = \frac{500}{f_{yk} \cdot \frac{A_{s,req}}{A_{s,prov}}} = "
                r"\frac{500}{500.000 \cdot \frac{200.000}{250.000}} = 1.250 \ -",
            ),
            ("short", r"\frac{310}{\sigma_s} = 1.250 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_yk = 500.0
        a_s_req = 200.0
        a_s_prov = 250.0

        # Object to test
        latex = Form7Dot1MultiplicationFactorLimitSlenderness(f_yk=f_yk, a_s_req=a_s_req, a_s_prov=a_s_prov).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.38 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA, N
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot38MaxShearStress(Formula):
    r"""Class representing formula 6.38 for the calculation of the maximum shear stress, [$v_{Ed}$]."""

    label = "6.38"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        beta: DIMENSIONLESS,
        v_ed: N,
        u_i: MM,
        d: MM,
    ) -> None:
        r"""[$v_{Ed}$] Maximum shear stress [$MPa$].

        EN 1992-1-1:2004 art.6.4.3(3) - Formula (6.38)

        Parameters
        ----------
        beta : DIMENSIONLESS
            [$\beta$] Factor which depends on the distribution of the support reaction, see equation 6.39.
        v_ed : N
            [$V_{Ed}$] Design value of the shear force [$N$].
        u_i : MM
            [$u_{i}$] Length of the control perimeter being considered [$mm$].
        d : MM
            [$d$] Mean effective depth of the slab, which may be taken as (dy + dz)/2 [$mm$].
        """
        super().__init__()
        self.beta = beta
        self.v_ed = v_ed
        self.u_i = u_i
        self.d = d

    @staticmethod
    def _evaluate(
        beta: DIMENSIONLESS,
        v_ed: N,
        u_i: MM,
        d: MM,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            beta=beta,
            v_ed=v_ed,
        )
        raise_if_less_or_equal_to_zero(u_i=u_i, d=d)

        return beta * v_ed / (u_i * d)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.38."""
        return LatexFormula(
            return_symbol=r"v_{Ed}",
            result=f"{self:.{n}f}",
            equation=r"\beta \cdot \frac{V_{Ed}}{u_{i} \cdot d}",
            numeric_equation=rf"{self.beta:.{n}f} \cdot \frac{{{self.v_ed:.{n}f}}}{{{self.u_i:.{n}f} \cdot {self.d:.{n}f}}}",
            comparison_operator_label="=",
            unit="MPa",
        )



"""Formula 6.7n from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.math_helpers import cot
from blueprints.type_alias import DEG
from blueprints.validations import raise_if_greater_than_90, raise_if_negative


class Form6Dot7nCheckCotTheta(Formula):
    r"""Class representing formula 6.7n for check of cotangent of theta."""

    label = "6.7n"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        theta: DEG,
    ) -> None:
        r"""Check if cotangent of theta is between 1 and 2.5.

        EN 1992-1-1:2004 art.6.2.3(2) - Formula (6.7n)

        Parameters
        ----------
        theta : DEG
            [$\theta$] angle between the concrete compression strut and the beam axis perpendicular to the shear force [$-$].
        """
        super().__init__()
        self.theta = theta

    @staticmethod
    def _evaluate(
        theta: DEG,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_greater_than_90(theta=theta)
        raise_if_negative(theta=theta)

        return (cot(theta) >= 1) and (cot(theta) <= 2.5)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.7n."""
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=r"1 \leq \cot(\theta) \leq 2.5",
            numeric_equation=rf"1 \leq \cot({self.theta:.{n}f}) \leq 2.5",
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula 6.40af of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_40af import (
    Form6Dot40afHollowSections,
    Form6Dot40afWeldedBoxSections,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot40afHollowSections:
    """Validation for formula 6.40af (hollow sections) from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 7000.0  # updated default value
        h = 300.0
        t = 10.0

        formula = Form6Dot40afHollowSections(a=a, h=h, t=t)
        manually_calculated_result = 0.14285714285  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "h", "t"),
        [
            (-7000.0, 300.0, 10.0),  # a is negative
            (7000.0, -300.0, 10.0),  # h is negative
            (7000.0, 300.0, -10.0),  # t is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, h: float, t: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot40afHollowSections(a=a, h=h, t=t)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"a_f = \min \left( \frac{A - 2 \cdot h \cdot t}{A}, 0.5 \right) = "
                r"\min \left( \frac{7000.000 - 2 \cdot 300.000 \cdot 10.000}{7000.000}, 0.5 \right) = 0.143 \ -",
            ),
            (
                "complete_with_units",
                r"a_f = \min \left( \frac{A - 2 \cdot h \cdot t}{A}, 0.5 \right) = "
                r"\min \left( \frac{7000.000 \ mm^2 - 2 \cdot 300.000 \ mm \cdot 10.000 \ mm}"
                r"{7000.000 \ mm^2}, 0.5 \right) = 0.143 \ -",
            ),
            ("short", r"a_f = 0.143 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 7000.0  # updated default value
        h = 300.0
        t = 10.0

        latex = Form6Dot40afHollowSections(a=a, h=h, t=t).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."


class TestForm6Dot40afWeldedBoxSections:
    """Validation for formula 6.40af (welded box sections) from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        a = 7000.0  # updated default value
        h = 300.0
        t_w = 10.0

        formula = Form6Dot40afWeldedBoxSections(a=a, h=h, t_w=t_w)
        manually_calculated_result = 0.14285714285  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "h", "t_w"),
        [
            (-7000.0, 300.0, 10.0),  # a is negative
            (7000.0, -300.0, 10.0),  # h is negative
            (7000.0, 300.0, -10.0),  # t_w is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, h: float, t_w: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot40afWeldedBoxSections(a=a, h=h, t_w=t_w)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"a_f = \min \left( \frac{A - 2 \cdot h \cdot t_w}{A}, 0.5 \right) = "
                r"\min \left( \frac{7000.000 - 2 \cdot 300.000 \cdot 10.000}{7000.000}, 0.5 \right) = 0.143 \ -",
            ),
            (
                "complete_with_units",
                r"a_f = \min \left( \frac{A - 2 \cdot h \cdot t_w}{A}, 0.5 \right) = "
                r"\min \left( \frac{7000.000 \ mm^2 - 2 \cdot 300.000 \ mm \cdot 10.000 \ mm}"
                r"{7000.000 \ mm^2}, 0.5 \right) = 0.143 \ -",
            ),
            ("short", r"a_f = 0.143 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        a = 7000.0  # updated default value
        h = 300.0
        t_w = 10.0

        latex = Form6Dot40afWeldedBoxSections(a=a, h=h, t_w=t_w).latex()

        actual = {
            "complete": latex.complete,
            "complete_with_units": latex.complete_with_units,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Module for concrete material properties."""

import math
import re
from dataclasses import dataclass, field
from enum import Enum

from blueprints.type_alias import DIMENSIONLESS, KG_M3, MM, MPA, PER_DEGREE, PER_MILLE, PERCENTAGE
from blueprints.unit_conversion import GPA_TO_MPA


class ConcreteAggregateType(Enum):
    """Enumeration of concrete aggregate types."""

    QUARTZITE = "Quartzite"
    LIMESTONE = "Limestone"
    SANDSTONE = "Sandstone"
    BASALT = "Basalt"


class CementClass(Enum):
    """Enumeration of cement class according to NEN-EN 1992-1-1 art.3.1.2 (6)."""

    S = "S"
    R = "R"
    N = "N"


class ConcreteStrengthClass(Enum):
    """Enumeration of concrete strength classes based on table 3.1 from NEN-EN 1992-1-1."""

    C12_15 = "C12/15"
    C16_20 = "C16/20"
    C20_25 = "C20/25"
    C25_30 = "C25/30"
    C30_37 = "C30/37"
    C35_45 = "C35/45"
    C40_50 = "C40/50"
    C45_55 = "C45/55"
    C50_60 = "C50/60"
    C55_67 = "C55/67"
    C60_75 = "C60/75"
    C70_85 = "C70/85"
    C80_95 = "C80/95"
    C90_105 = "C90/105"


class DiagramType(Enum):
    """Enumeration of diagram types of stress-strain relations."""

    BILINEAR = "Bi-Linear"
    PARABOLIC = "Parabolic"
    USER = "User defined"


class CementType(Enum):
    """Enumeration of CEM types."""

    CEM_I = "CEM I"
    CEM_II = "CEM II"
    CEM_III = "CEM III"
    CEM_III_A = "CEM IIIA"
    CEM_III_B = "CEM IIIB"
    CEM_III_C = "CEM IIIC"
    CEM_IV = "CEM IV"
    CEM_V = "CEM V"


@dataclass(frozen=True)
class ConcreteMaterial:
    r"""Representation of the strength and deformation characteristics for concrete material based on the analytical
    relation shown on table 3.1 from NEN-EN 1992-1-1.

    Parameters
    ----------
    concrete_class: ConcreteStrengthClass
        Enumeration of concrete strength classes (default: C30/37)
    cement_class: CementClass
        Cement class needed for calculation of Creep and shrinkage strain (Annex B from NEN-EN 1992-1-1). Classes are S, N and R. (default= N)
    density: KG_M3
        Unit mass of concrete [$kg/m^3$] (default= 2500.0 for regular reinforced concrete)
    diagram_type: DiagramType
        Type of stress-strain diagram (Figures 3.2, 3.3 and 3.4 from NEN-EN 1992-1-1) (default= Bi-Linear)
    aggregate_type: ConcreteAggregateType
        Type of aggregate in the concrete material (NEN-EN 1992-1-1) (default= Quartzite)
    aggregate_size: MM
        Largest nominal maximum aggregate size [$mm$] (NEN-EN 1992-1-1) (default= 16.0)
    use_plain_concrete_diagram: bool
        Use a Stress-strain diagram for plain or lightly reinforced concrete (NEN-EN 1992-1-1) (default= False)
    material_factor: DIMENSIONLESS
        Partial safety factor [$\gamma_c$] for concrete according to NEN-EN 1992-1-1 art.2.4.2.4 [$-$] (default= 1.5)
    thermal_coefficient: PER_DEGREE
        Thermal coefficient of the material [$1/°C$] (default= 1e-5)
    cement_type: CementType
        Type of cement in the concrete material (NEN-EN 1992-1-1) (default= CEM III)
    custom_name: str
        Use a custom name for the concrete material (default= concrete class name)
    custom_e_c: MPA
        Use a custom modulus of elasticity of concrete [$MPa$]
        If no custom value is given, the value [$E_{cm}$] from table 3.1 of NEN-EN 1992-1-1 is used.

    """

    concrete_class: ConcreteStrengthClass = field(default=ConcreteStrengthClass.C30_37)
    cement_class: CementClass = field(default=CementClass.N)
    density: KG_M3 = field(default=2500.0, metadata={"unit": "kg/m³"})
    diagram_type: DiagramType = field(default=DiagramType.BILINEAR)
    aggregate_type: ConcreteAggregateType = field(default=ConcreteAggregateType.QUARTZITE)
    aggregate_size: MM = field(default=16.0, metadata={"unit": "mm"})
    use_plain_concrete_diagram: bool = field(default=False)
    material_factor: DIMENSIONLESS = field(default=1.5)
    thermal_coefficient: PER_DEGREE = field(default=1e-5, metadata={"unit": "1/°C"})
    cement_type: CementType = field(default=CementType.CEM_III)
    custom_name: str | None = field(default=None, compare=False)
    custom_e_c: MPA | None = field(default=None, metadata={"unit": "MPa"})

    @property
    def name(self) -> str:
        """Name of the concrete material.

        Returns
        -------
        str
            Example: "C30/37"
        """
        if self.custom_name:
            return self.custom_name
        return self.concrete_class.value

    @property
    def e_c(self) -> MPA:
        r"""[$E_c$] Modulus of elasticity of concrete [$MPa$].

        If no custom value is given, the value [$E_{cm}$] from table 3.1 of NEN-EN 1992-1-1 is used.

        Returns
        -------
        MPA
            Example: 32836 (for C30/37)
        """
        if self.custom_e_c:
            return self.custom_e_c
        return self.e_cm

    @property
    def f_ck(self) -> MPA:
        r"""[$f_{ck}$] Characteristic compressive cylinder strength of concrete at 28 days [$MPa$].

        Returns
        -------
        MPA
            Example: 30 (for C30/37)
        """
        value = self.concrete_class.value
        if match := re.search(pattern=r"C(\d+)/", string=value):
            return int(match.group(1))
        raise ValueError("No match found for f_ck. Concrete class is invalid.")

    @property
    def f_ck_cube(self) -> MPA:
        r"""[$f_{ck,cube}$] Characteristic compressive cubic strength of concrete at 28 days [$MPa$].

        Returns
        -------
        MPA
            Example: 37 (for C30/37)
        """
        value = self.concrete_class.value
        if match := re.search(pattern=r"/(\d+)", string=value):
            return int(match.group(1))
        raise ValueError("No match found for f_ck_cube. Concrete class is invalid.")

    @property
    def f_cd(self) -> MPA:
        r"""[$f_{cd}$] Design value of concrete compressive strength (NEN-EN 1992-1-1 art.3.1.6 (1)) [$MPa$].

        Returns
        -------
        MPA
            Example: 20 (for C30/37)
        """
        return self.f_ck / self.material_factor

    @property
    def f_cm(self) -> MPA:
        r"""[$f_{cm}$] Mean value of concrete cylinder compressive strength [$MPa$].

        Returns
        -------
        MPA
            Example: 38 (for C30/37)
        """
        return self.f_ck + 8

    @property
    def f_cm_cube(self) -> MPA:
        r"""[$f_{cm,cube}$] Mean value of concrete cubic compressive strength [$MPa$].

        Returns
        -------
        MPA
            Example: 45 (for C30/37)
        """
        return self.f_ck_cube + 8

    @property
    def f_ctm(self) -> MPA:
        r"""[$f_{ctm}$] Mean value of axial tensile strength of concrete [$MPa$].

        Returns
        -------
        MPA
            Example: 2.896468153816889 (for C30/37)
        """
        if self.f_ck <= 50:
            return 0.30 * self.f_ck ** (2 / 3)
        return 2.12 * math.log(1 + (self.f_cm / 10))

    @property
    def sigma_cr(self) -> MPA:
        r"""[$\sigma_{cr}$] Crack tensile stress (long term) equal to [$0.6 \cdot f_{ctm}$] [$MPa$].

        Returns
        -------
        MPA
            Example: 1.7378808922901334 (for C30/37)
        """
        return self.f_ctm * 0.6

    @property
    def strain_cr(self) -> MPA:
        r"""[$\epsilon_{cr}$] Strain at crack tensile stress (long term) equal to [$\sigma_{cr} / E_{cm}$] [$MPa$].

        Returns
        -------
        MPA
            Example: 5.2926083941105295e-05 (for C30/37)
        """
        return self.sigma_cr / self.e_cm

    @property
    def f_ctk_0_05(self) -> MPA:
        r"""[$f_{ctk,0.05}$] Mean value of axial tensile strength of concrete, 5% fractile [$MPa$].

        Returns
        -------
        MPA
            Example: 2.027527707671822 (for C30/37)
        """
        return self.f_ctm * 0.7

    @property
    def f_ctd(self) -> MPA:
        r"""[$f_{ctd}$] Design value of tensile strength of concrete [$MPa$].

        Returns
        -------
        MPA
        Example: 1.3516851384478814 (for C30/37)
        """
        return self.f_ctk_0_05 / 1.5

    @property
    def f_ctk_0_95(self) -> MPA:
        r"""[$f_{ctk,0.95}$] Mean value of axial tensile strength of concrete, 95% fractile [$MPa$].

        Returns
        -------
        MPA
            Example: 3.765408599961956 (for C30/37)
        """
        return self.f_ctm * 1.3

    @property
    def e_cm(self) -> MPA:
        r"""[$E_{cm}$] Secant modulus of elasticity of concrete [$MPa$].

        Returns
        -------
        MPA
            Example: 32836 (for C30/37)
        """
        return int(22 * ((self.f_cm / 10) ** 0.3) * GPA_TO_MPA)

    @property
    def custom_e_c_present(self) -> bool:
        r"""Checks if the value of [$E_c$] equal is to the value of [$E_{cm}$] from table 3.1 of NEN-EN 1992-1-1.

        Returns
        -------
        bool
            Example: False if ConcreteMaterial.e_cm is equal to 32836 (for C30/37)
        """
        return self.e_c != self.e_cm

    @property
    def eps_c1(self) -> PER_MILLE:
        r"""[$\epsilon_{c1}$] Compressive strain in the concrete at the peak stress [$f_c$] [$‰$ (per mille)]. Value with a maximum of 2.8 ‰.
        Check Figure 3.2 of NEN-EN 1992-1-1.

        Returns
        -------
        PER_MILLE
            Example: 2.1618768697354804 (for C30/37)
        """
        return min(0.7 * self.f_cm**0.31, 2.8)

    @property
    def eps_cu1(self) -> PER_MILLE:
        r"""[$\epsilon_{cu1}$] Nominal ultimate compressive strain in the concrete [$‰$ (per mille)]. Check Figure 3.2 of NEN-EN 1992-1-1.

        Returns
        -------
        PER_MILLE
            Example: 3.5 (for C30/37)
        """
        if self.f_ck >= 50:
            return 2.8 + 27 * ((98 - self.f_cm) / 100) ** 4
        return 3.5

    @property
    def eps_c2(self) -> PER_MILLE:
        r"""[$\epsilon_{c2}$] Compressive strain at reaching the maximum strength according to Table 3.1 [$‰$ (per mille)].
        Check Figure 3.3 of NEN-EN 1992-1-1.

        Returns
        -------
        PER_MILLE
            Example: 2.0 (for C30/37)
        """
        if self.f_ck >= 50:
            return 2.0 + 0.085 * (self.f_ck - 50) ** 0.53
        return 2.0

    @property
    def eps_cu2(self) -> PER_MILLE:
        r"""[$\epsilon_{cu2}$] Nominal ultimate compressive strain in the concrete according to Table 3.1 [$‰$ (per mille)].
        Check Figure 3.3 of NEN-EN 1992-1-1.

        Returns
        -------
        PER_MILLE
            Example: 3.5 (for C30/37)
        """
        if self.f_ck >= 50:
            return 2.6 + 35 * ((90 - self.f_ck) / 100) ** 4
        return 3.5

    @property
    def n_factor(self) -> DIMENSIONLESS:
        r"""[$n$] factor from table 3.1 of NEN-EN 1992-1-1 [$-$].

        Returns
        -------
        DIMENSIONLESS
            Example: 2.0 (for C30/37)
        """
        if self.f_ck >= 50:
            return 1.4 + 23.4 * ((90 - self.f_ck) / 100) ** 4
        return 2.0

    @property
    def eps_c3(self) -> PER_MILLE:
        r"""[$\epsilon_{c3}$] Compressive strain at reaching the maximum strength according to Bi-linear stress-strain relation [$‰$ (per mille)].

        Check Figure 3.4 of NEN-EN 1992-1-1.

        Returns
        -------
        PER_MILLE
            Example: 1.75 (for C30/37)
        """
        if self.f_ck >= 50:
            return 1.75 + 0.55 * ((self.f_ck - 50) / 40)
        return 1.75

    @property
    def eps_cu3(self) -> PER_MILLE:
        r"""[$\epsilon_{cu3}$] Nominal ultimate compressive strain in concrete according to Bi-linear stress-strain relation [$‰$ (per mille)].

        Check Figure 3.4 of NEN-EN 1992-1-1.

        Returns
        -------
        PER_MILLE
            Example: 3.5 (for C30/37)
        """
        if self.f_ck >= 50:
            return 2.6 + 35 * ((90 - self.f_ck) / 100) ** 4
        return 3.5

    def rho_min(self, f_yd: MPA) -> PERCENTAGE:
        r"""[$\rho_{min}$] Minimum reinforcement ratio (CB2, 7de druk 2011, pag.55) [$%$].

        Parameters
        ----------
        f_yd: MPA
            Design yield strength of reinforcement [$MPa$]

        Returns
        -------
        PERCENTAGE
        """
        return (0.223 * (self.f_ctm / f_yd)) * 100



"""Formula 6.28 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.math_helpers import cot
from blueprints.type_alias import DEG, MM, MM2, MPA, NMM
from blueprints.validations import raise_if_greater_than_90, raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot28RequiredCrossSectionalArea(Formula):
    r"""Class representing formula 6.28 for the calculation of the required cross-sectional area of the longitudinal reinforcement.
    The description of the equation states that it is used to calculate the total reinforcement area. Therefore the
    calculation has been rewritten to find the solution to that question, [$\Sigma A_{sl}$].
    """

    label = "6.28"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        u_k: MM,
        f_yd: MPA,
        t_ed: NMM,
        a_k: MM2,
        theta: DEG,
    ) -> None:
        r"""[$\Sigma A_{sl}$] Required cross-sectional area of the longitudinal reinforcement [$mm^2$].

        EN 1992-1-1:2004 art.6.3.2(3) - Formula (6.28)

        Parameters
        ----------
        u_k : MM
            [$u_k$] Perimeter of the area A_k [$mm$].
        f_yd : MPA
            [$f_{yd}$] Design yield stress of the longitudinal reinforcement [$MPa$].
        t_ed : NMM
            [$T_{Ed}$] Design value of the torsional moment [$Nmm$].
        a_k : MM2
            [$A_k$] Area enclosed by the centre-lines of the connecting walls, including inner hollow areas [$mm^2$].
        theta : DEG
            [$\theta$] Angle of compression struts (see Figure 6.5) [$degrees$].
        """
        super().__init__()
        self.u_k = u_k
        self.f_yd = f_yd
        self.t_ed = t_ed
        self.a_k = a_k
        self.theta = theta

    @staticmethod
    def _evaluate(
        u_k: MM2,
        f_yd: MPA,
        t_ed: NMM,
        a_k: MM2,
        theta: DEG,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            u_k=u_k,
            t_ed=t_ed,
        )
        raise_if_less_or_equal_to_zero(f_yd=f_yd, a_k=a_k, theta=theta)
        raise_if_greater_than_90(theta=theta)

        return (u_k / f_yd) * (t_ed / (2 * a_k)) * cot(theta)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.28."""
        return LatexFormula(
            return_symbol=r"\Sigma A_{sl}",
            result=f"{self:.{n}f}",
            equation=r"\frac{u_k}{f_{yd}} \cdot \frac{T_{Ed}}{2 \cdot A_k} \cdot \cot(\theta)",
            numeric_equation=rf"\frac{{{self.u_k:.{n}f}}}{{{self.f_yd:.{n}f}}} \cdot \frac{{{self.t_ed:.{n}f}}}"
            rf"{{2 \cdot {self.a_k:.{n}f}}} \cdot \cot({self.theta:.{n}f})",
            comparison_operator_label="=",
            unit="mm^2",
        )



"""Formula 6.2 from EN 1993-1-1:2005: Chapter 6 - Ultimate limit state."""

# pylint: disable=arguments-differ
from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, KNM, RATIO
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot2UtilizationRatio(Formula):
    """Class representing form 6.2 for the calculation of the utilization ratio [$UC$]."""

    label = "6.2"
    source_document = EN_1993_1_1_2005

    def __init__(self, n_ed: KN, n_rd: KN, m_y_ed: KNM, m_y_rd: KNM, m_z_ed: KNM, m_z_rd: KNM) -> None:
        r"""
        [$UC$] The calculation of the utilization ratio [$-$].

        EN 1993-1-1:2005 art.6.2.1(7) - Formula (6.2)

        Parameters
        ----------
        n_ed : kN
            [$N_{Ed}$] Contains the design axial force [$kN$].
        n_rd : kN
            [$N_{Rd}$] Contains the design axial resistance [$kN$].
        m_y_ed : kNm
            [$M_{y,Ed}$] Contains the design moment about the y-axis [$kNm$].
        m_y_rd : kNm
            [$M_{y,Rd}$] Contains the design moment resistance about the y-axis [$kNm$].
        m_z_ed : kNm
            [$M_{z,Ed}$] Contains the design moment about the z-axis [$kNm$].
        m_z_rd : kNm
            [$M_{z,Rd}$] Contains the design moment resistance about the z-axis [$kNm$].

        Returns
        -------
        None
        """
        super().__init__()
        self.n_ed = n_ed
        self.n_rd = n_rd
        self.m_y_ed = m_y_ed
        self.m_y_rd = m_y_rd
        self.m_z_ed = m_z_ed
        self.m_z_rd = m_z_rd

    @staticmethod
    def _evaluate(
        n_ed: KN,
        n_rd: KN,
        m_y_ed: KNM,
        m_y_rd: KNM,
        m_z_ed: KNM,
        m_z_rd: KNM,
    ) -> RATIO:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(n_rd=n_rd, m_y_rd=m_y_rd, m_z_rd=m_z_rd)
        raise_if_negative(n_ed=n_ed, m_y_ed=m_y_ed, m_z_ed=m_z_ed)

        return (n_ed / n_rd) + (m_y_ed / m_y_rd) + (m_z_ed / m_z_rd)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for form 6.2."""
        numeric_equation = (
            rf"\frac{{{self.n_ed:.{n}f}}}{{{self.n_rd:.{n}f}}} + "
            rf"\frac{{{self.m_y_ed:.{n}f}}}{{{self.m_y_rd:.{n}f}}} + "
            rf"\frac{{{self.m_z_ed:.{n}f}}}{{{self.m_z_rd:.{n}f}}}"
        )
        return LatexFormula(
            return_symbol=r"UC",
            result=f"{self:.{n}f}",
            equation=r"\frac{N_{Ed}}{N_{Rd}} + \frac{M_{y,Ed}}{M_{y,Rd}} + \frac{M_{z,Ed}}{M_{z,Rd}}",
            numeric_equation=numeric_equation,
            comparison_operator_label="=",
        )



"""Formula 6.19 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot19CheckDesignElasticShearResistance(Formula):
    r"""Class representing formula 6.19 for checking the design elastic shear resistance."""

    label = "6.19"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        tau_ed: MPA,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
    ) -> None:
        r"""Check the design elastic shear resistance.

        EN 1993-1-1:2005 art.6.2.6(4) - Formula (6.19)

        Parameters
        ----------
        tau_ed : MPA
            [$\tau_{Ed}$] Design shear stress [MPa].
        f_y : MPA
            [$f_{y}$] Yield strength of the material [MPa].
        gamma_m0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial safety factor [-].
        """
        super().__init__()
        self.tau_ed = tau_ed
        self.f_y = f_y
        self.gamma_m0 = gamma_m0

    @staticmethod
    def _evaluate(
        tau_ed: MPA,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(gamma_m0=gamma_m0, f_y=f_y)
        raise_if_negative(tau_ed=tau_ed)

        return tau_ed / (f_y / (3**0.5 * gamma_m0)) <= 1.0

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.19."""
        _equation: str = r"\left( \frac{\tau_{Ed}}{f_{y} / (\sqrt{3} \cdot \gamma_{M0})} \leq 1.0 \right)"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\tau_{Ed}": f"{self.tau_ed:.{n}f}",
                r"f_{y}": f"{self.f_y:.{n}f}",
                r"\gamma_{M0}": f"{self.gamma_m0:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing formula A.1 from EN 1993-1-9:2005: Annex A - Determination of fatigue load parameters and verification formats."""

import pytest

from blueprints.codes.eurocode.en_1993_1_9_2005.annex_a_determination_of_fatigue_load_parameters_and_verification_formats.formula_a_2 import (
    FormADot2CriteriaBasedOnDamageAccumulation,
)
from blueprints.validations import NegativeValueError


class TestFormADot2CriteriaBasedOnDamageAccumulation:
    """Validation for formula A.2 from EN 1993-1-9:2005."""

    @pytest.mark.parametrize(
        ("d_d", "result_manual"),
        [
            (0.8, True),
            (1.1, False),
        ],
    )
    def test_evaluation(self, d_d: float, result_manual: bool) -> None:
        """Test the evaluation of the result."""
        assert FormADot2CriteriaBasedOnDamageAccumulation(d_d=d_d) == result_manual

    def test_raise_error_if_negative_n_e(self) -> None:
        """Test that a NegativeValueError is raised when a negative value is passed for n_E."""
        d_d = -0.1  # -
        with pytest.raises(NegativeValueError):
            FormADot2CriteriaBasedOnDamageAccumulation(d_d=d_d)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \rightarrow D_d \leq 1.0 \rightarrow 0.800 \leq 1.0 \rightarrow OK",
            ),
            ("short", r"CHECK \rightarrow OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        d_d = 0.8  # -

        # Object to test
        form_a_2_latex = FormADot2CriteriaBasedOnDamageAccumulation(d_d=d_d).latex()

        actual = {
            "complete": form_a_2_latex.complete,
            "short": form_a_2_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 5.23 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form5Dot23FactorConcreteStrengthClass(Formula):
    """Class representing formula 5.23 for the calculation of the factor for concrete strength class, [$k_{1}$]."""

    label = "5.23"
    source_document = EN_1992_1_1_2004

    def __init__(self, f_ck: MPA) -> None:
        r"""[$k_{1}$] Factor for concrete strength class.

        EN 1992-1-1:2004 art.5.8.6(3) - Formula (5.23)

        Parameters
        ----------
        f_ck : MPA
            [$f_{ck}$] is the characteristic compressive strength of concrete.
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(f_ck: MPA) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck)
        return (f_ck / 20) ** 0.5

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.23."""
        return LatexFormula(
            return_symbol=r"k_{1}",
            result=f"{self:.{n}f}",
            equation=r"\sqrt{\frac{f_{ck}}{20}}",
            numeric_equation=rf"\sqrt{{\frac{{{self.f_ck:.{n}f}}}{{20}}}}",
            comparison_operator_label="=",
        )



"""Testing formula 6.14 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_14 import (
    Form6Dot14MaxShearResistanceInclinedReinforcement,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot14MaxShearResistanceInclinedReinforcement:
    """Validation for formula 6.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        alpha_cw = 1.0  # dimensionless
        b_w = 300.0  # mm
        z = 400.0  # mm
        nu_1 = 0.6  # dimensionless
        f_cd = 30.0  # MPa
        theta = 30.0  # degrees
        alpha = 45.0  # degrees

        # Object to test
        formula = Form6Dot14MaxShearResistanceInclinedReinforcement(alpha_cw=alpha_cw, b_w=b_w, z=z, nu_1=nu_1, f_cd=f_cd, theta=theta, alpha=alpha)

        # Expected result, manually calculated
        manually_calculated_result = 1475307.43609  # N

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("alpha_cw", "b_w", "z", "nu_1", "f_cd", "theta", "alpha"),
        [
            (-1.0, 300.0, 400.0, 0.6, 30.0, 30.0, 45.0),  # alpha_cw is negative
            (1.0, -300.0, 400.0, 0.6, 30.0, 30.0, 45.0),  # b_w is negative
            (1.0, 300.0, -400.0, 0.6, 30.0, 30.0, 45.0),  # z is negative
            (1.0, 300.0, 400.0, -0.6, 30.0, 30.0, 45.0),  # nu_1 is negative
            (1.0, 300.0, 400.0, 0.6, -30.0, 30.0, 45.0),  # f_cd is negative
            (1.0, 300.0, 400.0, 0.6, 30.0, -30.0, 45.0),  # theta is negative
            (1.0, 300.0, 400.0, 0.6, 30.0, 30.0, -45.0),  # alpha is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(
        self, alpha_cw: float, b_w: float, z: float, nu_1: float, f_cd: float, theta: float, alpha: float
    ) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot14MaxShearResistanceInclinedReinforcement(alpha_cw=alpha_cw, b_w=b_w, z=z, nu_1=nu_1, f_cd=f_cd, theta=theta, alpha=alpha)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{Rd,max} = \alpha_{cw} \cdot b_{w} \cdot z \cdot \nu_{1} \cdot f_{cd} \cdot \frac{\cot(\theta) + "
                r"\cot(\alpha)}{1 + \cot^2(\theta)} = 1.000 \cdot 300.000 \cdot 400.000 \cdot 0.600 \cdot 30.000 \cdot "
                r"\frac{\cot(30.000) + \cot(45.000)}{1 + \cot^2(30.000)} = 1475307.436 \ N",
            ),
            ("short", r"V_{Rd,max} = 1475307.436 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        alpha_cw = 1.0  # dimensionless
        b_w = 300.0  # mm
        z = 400.0  # mm
        nu_1 = 0.6  # dimensionless
        f_cd = 30.0  # MPa
        theta = 30.0  # degrees
        alpha = 45.0  # degrees

        # Object to test
        latex = Form6Dot14MaxShearResistanceInclinedReinforcement(
            alpha_cw=alpha_cw, b_w=b_w, z=z, nu_1=nu_1, f_cd=f_cd, theta=theta, alpha=alpha
        ).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.38b of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_38b import Form5Dot38bCheckRelativeEccentricityRatio
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot38bCheckRelativeEccentricityRatio:
    """Validation for formula 5.38b from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        e_y = 30.0
        e_z = 4.0
        b_eq = 100.0
        h_eq = 101.0

        # Object to test
        formula = Form5Dot38bCheckRelativeEccentricityRatio(e_y=e_y, e_z=e_z, b_eq=b_eq, h_eq=h_eq)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    def test_evaluation_excentricity_zero(self) -> None:
        """Tests the evaluation of the result with e=0."""
        # Example values
        e_y = 30.0
        e_z = 0.0
        b_eq = 100.0
        h_eq = 101.0

        # Object to test
        formula = Form5Dot38bCheckRelativeEccentricityRatio(e_y=e_y, e_z=e_z, b_eq=b_eq, h_eq=h_eq)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("e_y", "e_z", "b_eq", "h_eq"),
        [
            (-30.0, 4.0, 100.0, 101.0),  # e_y is negative
            (30.0, -4.0, 100.0, 101.0),  # e_z is negative
            (30.0, 4.0, -100.0, 101.0),  # b_eq is negative
            (30.0, 4.0, 100.0, -101.0),  # h_eq is negative
            (30.0, 4.0, 0.0, 101.0),  # b_eq is zero
            (30.0, 4.0, 100.0, 0.0),  # h_eq is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, e_y: float, e_z: float, b_eq: float, h_eq: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form5Dot38bCheckRelativeEccentricityRatio(e_y=e_y, e_z=e_z, b_eq=b_eq, h_eq=h_eq)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to \left(\frac{e_{y}/h_{eq}}{e_{z}/b_{eq}} \leq 0.2 \text{ or } \frac{e_{z}/b_{eq}}{e_{y}/h_{eq}} \leq 0.2 \right) \to"
                r" \left(\frac{30.000/101.000}{4.000/100.000} \leq 0.2 \text{ or } \frac{4.000/100.000}{30.000/101.000} \leq 0.2 \right) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        e_y = 30.0
        e_z = 4.0
        b_eq = 100.0
        h_eq = 101.0

        # Object to test
        latex = Form5Dot38bCheckRelativeEccentricityRatio(e_y=e_y, e_z=e_z, b_eq=b_eq, h_eq=h_eq).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Test suite for CombinedSteelCrossSection."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_3_materials.table_3_1 import SteelStrengthClass
from blueprints.materials.steel import SteelMaterial
from blueprints.structural_sections.cross_section_rectangle import RectangularCrossSection
from blueprints.structural_sections.steel.steel_cross_sections._steel_cross_section import CombinedSteelCrossSection
from blueprints.structural_sections.steel.steel_element import SteelElement


class TestCombinedSteelCrossSection:
    """Test suite for CombinedSteelCrossSection."""

    def test_empty_combined_steel_cross_section(self, empty_combined_steel_cross_section: CombinedSteelCrossSection) -> None:
        """Test the code of the combined steel cross-section."""
        with pytest.raises(ValueError):
            _ = empty_combined_steel_cross_section.polygon

    def test_invalid_combined_elements(self, empty_combined_steel_cross_section: CombinedSteelCrossSection) -> None:
        """Test the code of the combined steel cross-section."""
        steel_material = SteelMaterial(steel_class=SteelStrengthClass.S355)
        empty_combined_steel_cross_section.elements = [
            SteelElement(
                cross_section=RectangularCrossSection(width=500, height=500, x=0, y=0),
                material=steel_material,
                nominal_thickness=500,
            ),
            SteelElement(
                cross_section=RectangularCrossSection(width=250, height=250, x=1000, y=0),
                material=steel_material,
                nominal_thickness=250,
            ),
        ]

        with pytest.raises(TypeError):
            _ = empty_combined_steel_cross_section.polygon



"""Testing formula 7.18 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_18 import Form7Dot18DeformationParameter
from blueprints.validations import NegativeValueError


class TestForm7Dot18DeformationParameter:
    """Validation for formula 7.18 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        zeta = 1 / 3.0
        alpha_ll = 1.4
        alpha_l = 0.8

        # Object to test
        formula = Form7Dot18DeformationParameter(zeta=zeta, alpha_ll=alpha_ll, alpha_l=alpha_l)

        # Expected result, manually calculated
        manually_calculated_result = 1.0  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("zeta", "alpha_ll", "alpha_l"),
        [
            (-0.5, 1.2, 0.8),  # zeta is negative
            (0.5, -1.2, 0.8),  # alpha_ll is negative
            (0.5, 1.2, -0.8),  # alpha_l is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, zeta: float, alpha_ll: float, alpha_l: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form7Dot18DeformationParameter(zeta=zeta, alpha_ll=alpha_ll, alpha_l=alpha_l)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\alpha = \zeta \cdot \alpha_{II} + (1 - \zeta) \cdot \alpha_{I} = "
                r"0.500 \cdot 1.200 + (1 - 0.500) \cdot 0.800 = 1.000 \ -",
            ),
            ("short", r"\alpha = 1.000 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        zeta = 0.5
        alpha_ll = 1.2
        alpha_l = 0.8

        # Object to test
        latex = Form7Dot18DeformationParameter(zeta=zeta, alpha_ll=alpha_ll, alpha_l=alpha_l).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 12.4 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_12_plain_and_lightly_reinforced_concrete_structures.formula_12_4 import (
    Form12Dot4PlainConcreteShearStress,
    Form12Dot4PlainConcreteShearStressComparison,
)
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm12Dot4PlainConcreteShearStress:
    """Validation for formula 12.4 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        k = 1.5  # Nationally determined parameter
        v_ed = 100000.0  # N
        a_cc = 50000.0  # mm^2

        # Object to test
        form_12_4 = Form12Dot4PlainConcreteShearStress(k=k, v_ed=v_ed, a_cc=a_cc)

        # Expected result, manually calculated
        manually_calculated_result = 3.0  # MPa

        assert round(form_12_4, 3) == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("k", "v_ed", "a_cc"),
        [
            (1.5, 100000.0, 0.0),
        ],
    )
    def test_raise_error_when_values_are_less_or_equal_to_zero(
        self,
        k: float,
        v_ed: float,
        a_cc: float,
    ) -> None:
        """Test values less or equal to zero for v_ed and a_cc."""
        with pytest.raises(LessOrEqualToZeroError):
            Form12Dot4PlainConcreteShearStress(k=k, v_ed=v_ed, a_cc=a_cc)

    @pytest.mark.parametrize(
        ("k", "v_ed", "a_cc"),
        [
            (1.5, -100000.0, 50000.0),
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self,
        k: float,
        v_ed: float,
        a_cc: float,
    ) -> None:
        """Test negative values for v_ed and a_cc."""
        with pytest.raises(NegativeValueError):
            Form12Dot4PlainConcreteShearStress(k=k, v_ed=v_ed, a_cc=a_cc)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\tau_{cp} = k \cdot \frac{V_{Ed}}{A_{cc}} = 1.500 \cdot \frac{100000.000}{50000.000} = 3.000",
            ),
            ("short", r"\tau_{cp} = 3.000"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        k = 1.5  # Nationally determined parameter
        v_ed = 100000.0  # N
        a_cc = 50000.0  # mm^2

        # Object to test
        form_12_4_latex = Form12Dot4PlainConcreteShearStress(
            k=k,
            v_ed=v_ed,
            a_cc=a_cc,
        ).latex()

        actual = {
            "complete": form_12_4_latex.complete,
            "short": form_12_4_latex.short,
            "string": str(form_12_4_latex),
        }

        assert actual[representation] == expected, f"{representation} representation failed."


class TestForm12Dot4PlainConcreteShearStressComparison:
    """Validation for Form12Dot4PlainConcreteShearStressComparison."""

    def test_comparison_true(self) -> None:
        """Test the comparison when sigma_cp is less than or equal to sigma_c_lim."""
        comparison = Form12Dot4PlainConcreteShearStressComparison(sigma_cp=1.0, sigma_c_lim=1.5)
        assert comparison.comparison is True

    def test_comparison_false(self) -> None:
        """Test the comparison when sigma_cp is greater than sigma_c_lim."""
        comparison = Form12Dot4PlainConcreteShearStressComparison(sigma_cp=2.0, sigma_c_lim=1.5)
        assert comparison.comparison is False

    @pytest.mark.parametrize(
        ("sigma_cp", "sigma_c_lim"),
        [
            (-1.0, 1.5),
            (1.0, -1.5),
        ],
    )
    def test_raise_error_when_negative_values_are_given(
        self,
        sigma_cp: float,
        sigma_c_lim: float,
    ) -> None:
        """Test negative values for sigma_cp and sigma_c_lim."""
        with pytest.raises(NegativeValueError):
            Form12Dot4PlainConcreteShearStressComparison(sigma_cp=sigma_cp, sigma_c_lim=sigma_c_lim)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \rightarrow \sigma_{cp} ≤ \sigma_{c,lim} \rightarrow 1.000 ≤ 1.500 \rightarrow OK",
            ),
            (
                "complete_not_ok",
                r"CHECK \rightarrow \sigma_{cp} ≤ \sigma_{c,lim} \rightarrow 2.000 ≤ 1.500 \rightarrow \text{Not OK}",
            ),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the comparison."""
        if representation == "complete":
            comparison = Form12Dot4PlainConcreteShearStressComparison(sigma_cp=1.0, sigma_c_lim=1.5)
        else:
            comparison = Form12Dot4PlainConcreteShearStressComparison(sigma_cp=2.0, sigma_c_lim=1.5)

        actual = {
            "complete": comparison.latex().complete,
            "complete_not_ok": comparison.latex().complete,
        }

        assert actual[representation] == expected, f"{representation} representation failed."

    def test_str(self) -> None:
        """Test the string representation of the comparison."""
        comparison = Form12Dot4PlainConcreteShearStressComparison(sigma_cp=1.0, sigma_c_lim=1.5)
        expected_str = r"CHECK \rightarrow \sigma_{cp} ≤ \sigma_{c,lim} \rightarrow 1.000 ≤ 1.500 \rightarrow OK"
        assert str(comparison) == expected_str

        comparison_not_ok = Form12Dot4PlainConcreteShearStressComparison(sigma_cp=2.0, sigma_c_lim=1.5)
        expected_str_not_ok = r"CHECK \rightarrow \sigma_{cp} ≤ \sigma_{c,lim} \rightarrow 2.000 ≤ 1.500 \rightarrow " r"\text{Not OK}"
        assert str(comparison_not_ok) == expected_str_not_ok



"""Testing formula 8.20 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_20 import (
    Form8Dot20BondStrengthAnchorageULS,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot20BondStrengthAnchorageULS:
    """Validation for formula 8.20 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # example values
        eta_p2 = 1.4
        eta_1 = 1.0
        f_ctd = 2.5  # MPa
        form_8_20 = Form8Dot20BondStrengthAnchorageULS(eta_p2=eta_p2, eta_1=eta_1, f_ctd=f_ctd)
        # manually calculated result
        manually_calculated_result = 3.5  # MPa

        assert form_8_20 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_if_negative_values(self) -> None:
        """Test that a NegativeValueError is raised when negative values are passed."""
        eta_p2 = -1.4
        eta_1 = 1.0
        f_ctd = 2.5  # MPa
        with pytest.raises(NegativeValueError):
            Form8Dot20BondStrengthAnchorageULS(eta_p2=eta_p2, eta_1=eta_1, f_ctd=f_ctd)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            ("complete", r"f_{bpd} = \eta_{p2} \cdot \eta_{1} \cdot f_{ctd} = 1.400 \cdot 1.000 \cdot 2.500 = 3.500"),
            ("short", r"f_{bpd} = 3.500"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # example values
        eta_p2 = 1.4
        eta_1 = 1.0
        f_ctd = 2.5  # MPa
        form_8_20 = Form8Dot20BondStrengthAnchorageULS(eta_p2=eta_p2, eta_1=eta_1, f_ctd=f_ctd)

        # Object to test
        form_8_20_latex = form_8_20.latex()

        actual = {"complete": form_8_20_latex.complete, "short": form_8_20_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 4.4N from EN 1992-1-1:2004: Chapter 4 - Durability and cover to reinforcement."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import MM
from blueprints.validations import raise_if_less_or_equal_to_zero


class Form4Dot4NCheckExecutionTolerances(Formula):
    r"""Class representing formula 4.4N for calculating the allowance in design for execution tolerances
    [$\Delta c_{dev}$] [$mm$]. Used, where it can be assured that a very accurate measurement device is used
    for monitoring and non conforming members are rejected (e.g. precast elements).

    EN 1992-1-1:2004 art.4.4.1.3 (3) - formula (4.4N)

    Parameters
    ----------
    delta_cdev : MM
        [$\Delta c_{dev}$] Concrete cover including execution tolerances [$mm$].
    """

    label = "4.4N"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        delta_cdev: MM,
    ) -> None:
        super().__init__()
        self.delta_cdev = delta_cdev

    @staticmethod
    def _evaluate(
        delta_cdev: MM,
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(delta_cdev=delta_cdev)

        return 0 <= delta_cdev <= 10

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 4.3N."""
        _equation: str = r"0 \leq \Delta c_{dev} \leq 10"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\Delta c_{dev}": f"{self.delta_cdev:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Testing sub-formulas for 8.2 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_2 import (
    SubForm8Dot2CoefficientQualityOfBond,
)


class TestSubForm8Dot2CoefficientQualityOfBond:
    """Validation for sub-formula 8.2 from EN 1992-1-1:2004."""

    def test_evaluation_good(self) -> None:
        """Test the evaluation of the result."""
        bond_quality = "good"  # str
        sub_form_8_2_1 = SubForm8Dot2CoefficientQualityOfBond(bond_quality=bond_quality)
        # Expected result, manually calculated
        manually_result = 1
        assert sub_form_8_2_1 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_evaluation_other(self) -> None:
        """Test the evaluation of the result."""
        bond_quality = "other"  # str
        sub_form_8_2_2 = SubForm8Dot2CoefficientQualityOfBond(bond_quality=bond_quality)
        # Expected result, manually calculated
        manually_result = 0.7
        assert sub_form_8_2_2 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_raise_error_when_invalid_bond_quality_is_given(self) -> None:
        """Test an invalid bond quality."""
        # Example values
        bond_quality = "military_grade"  # str

        with pytest.raises(ValueError):
            SubForm8Dot2CoefficientQualityOfBond(bond_quality=bond_quality)



"""Testing formula 7.4 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_4 import Form7Dot4MeanStressConcrete
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot4MeanStressConcrete:
    """Validation for formula 7.4 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        n_ed = 1000.0  # N
        b = 200.0  # mm
        h = 300.0  # mm

        # Object to test
        formula = Form7Dot4MeanStressConcrete(n_ed=n_ed, b=b, h=h)

        # Expected result, manually calculated
        manually_calculated_result = 1000 / 200 / 300  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("n_ed", "b", "h"),
        [
            (-1000.0, 200.0, 300.0),  # n_ed is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given_negative(self, n_ed: float, b: float, h: float) -> None:
        """Test invalid values for negative n_ed."""
        with pytest.raises(NegativeValueError):
            Form7Dot4MeanStressConcrete(n_ed=n_ed, b=b, h=h)

    @pytest.mark.parametrize(
        ("n_ed", "b", "h"),
        [
            (1000.0, 0.0, 300.0),  # b is zero
            (1000.0, 200.0, 0.0),  # h is zero
            (1000.0, -200.0, 300.0),  # b is negative
            (1000.0, 200.0, -300.0),  # h is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given_zero_or_negative(self, n_ed: float, b: float, h: float) -> None:
        """Test invalid values for zero or negative b and h."""
        with pytest.raises(LessOrEqualToZeroError):
            Form7Dot4MeanStressConcrete(n_ed=n_ed, b=b, h=h)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\sigma_c = \frac{N_{Ed}}{b \cdot h} = \frac{1000.000}{200.000 \cdot 300.000} = 0.017 \ MPa",
            ),
            ("short", r"\sigma_c = 0.017 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        n_ed = 1000.0
        b = 200.0
        h = 300.0

        # Object to test
        latex = Form7Dot4MeanStressConcrete(n_ed=n_ed, b=b, h=h).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 5.1 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, M
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form5Dot1Imperfections(Formula):
    r"""Class representing formula 5.1 for the calculation of initial inclination imperfections, [$\Theta_i$]."""

    label = "5.1"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        theta_0: DIMENSIONLESS,
        alpha_h: DIMENSIONLESS,
        alpha_m: DIMENSIONLESS,
    ) -> None:
        r"""[$\\Theta_i$] Initial inclination imperfections, [$\\Theta_i$], is a ratio between height and inclination of the member [-].

        EN 1992-1-1:2004 art.5.2(5) - Formula (5.1)

        Parameters
        ----------
        theta_0 : float
            [$\\Theta_0$] Basic value [-].
            Note: The value of [$\\Theta_0$] for use in a Country may be found in its National Annex.
            The recommended value is 1/200
        alpha_h : float
            [$\alpha_h$] Reduction factor for length or height [-].
            Use your own implementation of this value or use the SubForm5Dot1ReductionFactorLengthOrHeight class.
        alpha_m : float
            [$\alpha_m$] Reduction factor for number of members [-].
            Use your own implementation of this value or use the SubForm5Dot1ReductionFactorNumberOfMembers class.
        """
        super().__init__()
        self.theta_0 = theta_0
        self.alpha_h = alpha_h
        self.alpha_m = alpha_m

    @staticmethod
    def _evaluate(
        theta_0: DIMENSIONLESS,
        alpha_h: DIMENSIONLESS,
        alpha_m: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(theta_0=theta_0, alpha_h=alpha_h, alpha_m=alpha_m)
        return theta_0 * alpha_h * alpha_m

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.1."""
        return LatexFormula(
            return_symbol=r"\theta_i",
            result=f"{self:.4f}",
            equation=r"\theta_0 \cdot \alpha_h \cdot \alpha_m",
            numeric_equation=rf"{self.theta_0:.{n}f} \cdot {self.alpha_h:.{n}f} \cdot {self.alpha_m:.{n}f}",
            comparison_operator_label="=",
        )


class SubForm5Dot1ReductionFactorLengthOrHeight(Formula):
    r"""Class representing sub-formula 5.1 for the calculation of the reduction factor for length or height, [$\alpha_h$]."""

    label = "5.1"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        length: M,
    ) -> None:
        r"""[$\alpha_h$] Reduction factor for length or height [-].

        The calculated value of [$\alpha_h$] is between 2/3 and 1.0.

        EN 1992-1-1:2004 art.5.2(5) - Formula (5.1)

        Parameters
        ----------
        length : M
            [$\text{length}$] Length or height, see art.5.2(6) [$m$].
        """
        super().__init__()
        self.length = length

    @staticmethod
    def _evaluate(
        length: M,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(length=length)
        # the value of alpha_h is between 2/3 and 1.0
        alpha_h = 2 / np.sqrt(length)
        if alpha_h < 2 / 3:
            return 2 / 3
        if alpha_h > 1:
            return 1
        return alpha_h

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.1 subformula 1."""
        return LatexFormula(
            return_symbol=r"\alpha_h",
            result=f"{self:.{n}f}",
            equation=r"\min( \max(2 / \sqrt{l}; 2/3); 1)",
            numeric_equation=rf"\min( \max(2 / \sqrt{{{self.length:.{n}f}}}; 2/3); 1)",
            comparison_operator_label="=",
        )


class SubForm5Dot1ReductionFactorNumberOfMembers(Formula):
    r"""Class representing sub-formula 5.1 for the calculation of the reduction factor for number of members, [$\alpha_m$]."""

    label = "5.1"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        members: int,
    ) -> None:
        r"""[$\alpha_m$] Reduction factor for number of members [-].

        EN 1992-1-1:2004 art.5.2(5) - Formula (5.1)

        Parameters
        ----------
        members : int
            [$m$] Number of vertical members contributing to the total effect [-].
        """
        super().__init__()
        self.members = members

    @staticmethod
    def _evaluate(
        members: int,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_less_or_equal_to_zero(members=members)
        # the value of alpha_m is between 1.0 and 1.5
        return np.sqrt(0.5 * (1 + 1 / members))

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.1 subformula 2."""
        return LatexFormula(
            return_symbol=r"\alpha_m",
            result=f"{self:.{n}f}",
            equation=r"\sqrt{0.5 \cdot ( 1 + 1 / m)}",
            numeric_equation=rf"\sqrt{{0.5 \cdot ( 1 + 1 / {self.members:.{n}f})}}",
            comparison_operator_label="=",
        )



"""Table 3.1 of EN 1993-1-1:2005."""

from dataclasses import dataclass, field
from enum import Enum
from typing import ClassVar

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.type_alias import MM, MPA


class SteelStandardGroup(Enum):
    """Enumeration for steel standard groups in EN standards."""

    EN_10025_2 = "NEN-EN 10025-2"
    EN_10025_3 = "NEN-EN 10025-3"
    EN_10025_4 = "NEN-EN 10025-4"
    EN_10025_5 = "NEN-EN 10025-5"
    EN_10025_6 = "NEN-EN 10025-6"
    EN_10210_1 = "NEN-EN 10210-1"
    EN_10219_1 = "NEN-EN 10219-1"


class SteelStrengthClass(Enum):
    """Enumeration for steel strength classes with their standard group."""

    S235 = (SteelStandardGroup.EN_10025_2, "S 235")
    S275 = (SteelStandardGroup.EN_10025_2, "S 275")
    S355 = (SteelStandardGroup.EN_10025_2, "S 355")
    S450 = (SteelStandardGroup.EN_10025_2, "S 450")

    S275_N_NL_10025_3 = (SteelStandardGroup.EN_10025_3, "S 275 N/NL")
    S355_N_NL_10025_3 = (SteelStandardGroup.EN_10025_3, "S 355 N/NL")
    S420_N_NL_10025_3 = (SteelStandardGroup.EN_10025_3, "S 420 N/NL")
    S460_N_NL_10025_3 = (SteelStandardGroup.EN_10025_3, "S 460 N/NL")

    S275_M_ML_10025_4 = (SteelStandardGroup.EN_10025_4, "S 275 M/ML")
    S355_M_ML_10025_4 = (SteelStandardGroup.EN_10025_4, "S 355 M/ML")
    S420_M_ML_10025_4 = (SteelStandardGroup.EN_10025_4, "S 420 M/ML")
    S460_M_ML_10025_4 = (SteelStandardGroup.EN_10025_4, "S 460 M/ML")

    S235_W_10025_5 = (SteelStandardGroup.EN_10025_5, "S 235 W")
    S355_W_10025_5 = (SteelStandardGroup.EN_10025_5, "S 355 W")

    S460_Q_QL_QL1_10025_6 = (SteelStandardGroup.EN_10025_6, "S 460 Q/QL/QL1")

    S235_H_10210_1 = (SteelStandardGroup.EN_10210_1, "S 235 H")
    S275_H_10210_1 = (SteelStandardGroup.EN_10210_1, "S 275 H")
    S355_H_10210_1 = (SteelStandardGroup.EN_10210_1, "S 355 H")
    S275_NH_NLH_10210_1 = (SteelStandardGroup.EN_10210_1, "S 275 NH/NLH")
    S355_NH_NLH_10210_1 = (SteelStandardGroup.EN_10210_1, "S 355 NH/NLH")
    S420_NH_NLH_10210_1 = (SteelStandardGroup.EN_10210_1, "S 420 NH/NLH")
    S460_NH_NLH_10210_1 = (SteelStandardGroup.EN_10210_1, "S 460 NH/NLH")

    S235_H_10219_1 = (SteelStandardGroup.EN_10219_1, "S 235 H")
    S275_H_10219_1 = (SteelStandardGroup.EN_10219_1, "S 275 H")
    S355_H_10219_1 = (SteelStandardGroup.EN_10219_1, "S 355 H")
    S275_NH_NLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 275 NH/NLH")
    S355_NH_NLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 355 NH/NLH")
    S460_NH_NLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 460 NH/NLH")
    S275_MH_MLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 275 MH/MLH")
    S355_MH_MLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 355 MH/MLH")
    S420_MH_MLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 420 MH/MLH")
    S460_MH_MLH_10219_1 = (SteelStandardGroup.EN_10219_1, "S 460 MH/MLH")

    def __init__(self, standard_group: SteelStandardGroup, display_name: str) -> None:
        self.standard_group = standard_group
        self.display_name = display_name


@dataclass(frozen=True)
class Table3Dot1NominalValuesHotRolledStructuralSteel:
    """Implementation of table 3.1 from EN 1993-1-1:2005.

    Nominal values for yield strength (fy) and ultimate tensile strength (fu) for hot-rolled structural steel.

    Parameters
    ----------
    steel_class : SteelStrengthClass
        The steel strength class according to EN standards
    thickness : MM
        The nominal thickness of the steel element in mm

    Methods
    -------
    fy : MPA
        Returns the yield strength in N/mm². In the case of EN 10219-1, the values are not specified
        for thickness > 40 mm and will return None.
    fu : MPA
        Returns the ultimate tensile strength in N/mm². In the case of EN 10219-1, the values are not specified
        for thickness > 40 mm and will return None.

    Raises
    ------
    ValueError
        If an invalid steel class is provided
        If the thickness is not a positive number

    Examples
    --------
    >>> table = Table3Dot1NominalValuesHotRolledStructuralSteel(SteelStrengthClass.EN_10025_2_S355, 30)
    >>> table.fy
    355
    >>> table.fu
    490
    """

    steel_class: SteelStrengthClass
    thickness: MM
    label: str = field(init=False, default="Table 3.1")
    source_document: str = field(init=False, default=EN_1993_1_1_2005)

    # Class variable containing all strength data
    # Format: {SteelStrengthClass: (fy_t≤40, fu_t≤40, fy_t>40, fu_t>40)}
    _strength_data: ClassVar[dict[SteelStrengthClass, tuple[int, int, int | None, int | None]]] = {
        # EN 10025-2
        SteelStrengthClass.S235: (235, 360, 215, 360),
        SteelStrengthClass.S275: (275, 430, 255, 410),
        SteelStrengthClass.S355: (355, 490, 335, 470),
        SteelStrengthClass.S450: (440, 550, 410, 550),
        # EN 10025-3
        SteelStrengthClass.S275_N_NL_10025_3: (275, 390, 255, 370),
        SteelStrengthClass.S355_N_NL_10025_3: (355, 490, 335, 470),
        SteelStrengthClass.S420_N_NL_10025_3: (420, 520, 390, 520),
        SteelStrengthClass.S460_N_NL_10025_3: (460, 540, 430, 540),
        # EN 10025-4
        SteelStrengthClass.S275_M_ML_10025_4: (275, 370, 255, 360),
        SteelStrengthClass.S355_M_ML_10025_4: (355, 470, 335, 450),
        SteelStrengthClass.S420_M_ML_10025_4: (420, 520, 390, 500),
        SteelStrengthClass.S460_M_ML_10025_4: (460, 540, 430, 530),
        # EN 10025-5
        SteelStrengthClass.S235_W_10025_5: (235, 360, 215, 340),
        SteelStrengthClass.S355_W_10025_5: (355, 490, 335, 490),
        # EN 10025-6
        SteelStrengthClass.S460_Q_QL_QL1_10025_6: (460, 570, 440, 550),
        # EN 10210-1
        SteelStrengthClass.S235_H_10210_1: (235, 360, 215, 340),
        SteelStrengthClass.S275_H_10210_1: (275, 430, 255, 410),
        SteelStrengthClass.S355_H_10210_1: (355, 510, 335, 490),
        SteelStrengthClass.S275_NH_NLH_10210_1: (275, 390, 255, 370),
        SteelStrengthClass.S355_NH_NLH_10210_1: (355, 490, 335, 470),
        SteelStrengthClass.S420_NH_NLH_10210_1: (420, 540, 390, 520),
        SteelStrengthClass.S460_NH_NLH_10210_1: (460, 560, 430, 550),
        # EN 10219-1
        SteelStrengthClass.S235_H_10219_1: (235, 360, None, None),
        SteelStrengthClass.S275_H_10219_1: (275, 430, None, None),
        SteelStrengthClass.S355_H_10219_1: (355, 510, None, None),
        SteelStrengthClass.S275_NH_NLH_10219_1: (275, 370, None, None),
        SteelStrengthClass.S355_NH_NLH_10219_1: (355, 470, None, None),
        SteelStrengthClass.S460_NH_NLH_10219_1: (460, 550, None, None),
        SteelStrengthClass.S275_MH_MLH_10219_1: (275, 360, None, None),
        SteelStrengthClass.S355_MH_MLH_10219_1: (355, 470, None, None),
        SteelStrengthClass.S420_MH_MLH_10219_1: (420, 500, None, None),
        SteelStrengthClass.S460_MH_MLH_10219_1: (460, 530, None, None),
    }

    def __post_init__(self) -> None:
        """
        Validate the input parameters after initialization.

        Raises
        ------
        ValueError
            If the steel class is not in the strength data dictionary
            If the thickness is not a positive number
        """
        # Check if steel class is valid
        if self.steel_class not in self._strength_data:
            valid_classes = ", ".join([cls.name for cls in SteelStrengthClass])
            error_msg = f"Invalid steel class: {self.steel_class}. Valid classes are: {valid_classes}"
            raise ValueError(error_msg)

        # Check if thickness is positive
        if not isinstance(self.thickness, int | float) or self.thickness <= 0:
            raise ValueError(f"Thickness must be a positive number, got {self.thickness}")

    @property
    def fy(self) -> MPA:
        """
        Get the yield strength (fy) for the steel class and thickness.

        Returns
        -------
        MPA
            The yield strength in N/mm²

        Raises
        ------
        ValueError
            If the thickness is greater than 80 mm
        """
        # Check if thickness is within valid range (≤80 mm)
        if self.thickness > 80:
            raise ValueError(f"Thickness {self.thickness} mm exceeds maximum supported value of 80 mm")

        strength_values = self._strength_data[self.steel_class]

        # Choose value based on thickness (≤40 mm or >40 mm and ≤80 mm)
        result = strength_values[0] if self.thickness <= 40 else strength_values[2]

        if result is None:
            raise ValueError(
                f"Yield strength not specified for thickness > 40 mm for steel class '{self.steel_class.value}'. "
                f"Check {self.label} from {self.source_document}."
            )

        return result

    @property
    def fu(self) -> MPA:
        """
        Get the ultimate tensile strength (fu) for the steel class and thickness.

        Returns
        -------
        MPA
            The ultimate tensile strength in N/mm²

        Raises
        ------
        ValueError
            If the thickness is greater than 80 mm
        """
        # Check if thickness is within valid range (≤80 mm)
        if self.thickness > 80:
            raise ValueError(f"Thickness {self.thickness} mm exceeds maximum supported value of 80 mm")

        strength_values = self._strength_data[self.steel_class]

        # Choose value based on thickness (≤40 mm or >40 mm and ≤80 mm)
        result = strength_values[1] if self.thickness <= 40 else strength_values[3]

        if result is None:
            raise ValueError(
                f"Yield strength not specified for thickness > 40 mm for steel class '{self.steel_class.value}'. "
                f"Check {self.label} from {self.source_document}."
            )

        return result

    def __str__(self) -> str:
        """
        Return a string representation of the steel properties.

        Returns
        -------
        str
            String representation with class, thickness, and strength properties
        """
        return f"{self.steel_class.display_name}, t={round(self.thickness, 1)} mm, fy={round(self.fy, 2)} N/mm², fu={round(self.fu, 2)} N/mm²"



"""Formula 6.10a/bN from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form6Dot10abNStrengthReductionFactor(Formula):
    r"""Class representing formula 6.10a/bN for the calculation of the strength reduction factor for concrete cracked in shear."""

    label = "6.10a/bN"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ck: MPA,
    ) -> None:
        r"""[$\nu_{1}$] Strength reduction factor for concrete cracked in shear [$-$].

        EN 1992-1-1:2004 art.6.2.3(3) - Formula (6.10.aN and 6.10.bN)

        Parameters
        ----------
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        f_ck: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck)

        if f_ck <= 60:
            return 0.6
        return max(0.9 - f_ck / 200, 0.5)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.10a/bN."""
        return LatexFormula(
            return_symbol=r"\nu_{1}",
            result=f"{self:.{n}f}",
            equation=r"\begin{cases} 0.600 & \text{if } f_{ck} \leq 60 MPa \\ \max\left(0.9 - \frac{f_{ck}}{200}, 0.5\right) "
            r"& \text{if } f_{ck} /ge 60 MPa \end{cases}",
            numeric_equation=rf"\begin{{cases}} 0.600 & \text{{if }} {self.f_ck} \leq 60 MPa \\ "
            rf"\max\left(0.9 - \frac{{{self.f_ck}}}{{200}}, 0.5\right) & \text{{if }} {self.f_ck} /ge 60 MPa \end{{cases}}",
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 3.6 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_6 import Form3Dot6CreepDeformationOfConcrete


class TestForm3Dot6CreepDeformationOfConcrete:
    """Validation for formula 3.6 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = 0.34  # -
        sigma_c = 0.75  # MPa
        e_c = 2.45  # MPa
        form_3_6 = Form3Dot6CreepDeformationOfConcrete(phi_inf_t0=phi_inf_t0, sigma_c=sigma_c, e_c=e_c)

        # Expected result, manually calculated
        manually_calculated_result = 0.1040816

        assert form_3_6 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_phi_inf_t0_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = -0.34  # -
        sigma_c = 0.75  # MPa
        e_c = 2.45  # MPa

        with pytest.raises(ValueError):
            Form3Dot6CreepDeformationOfConcrete(phi_inf_t0=phi_inf_t0, sigma_c=sigma_c, e_c=e_c)

    def test_raise_error_when_negative_sigma_c_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = 0.34  # -
        sigma_c = -0.75  # MPa
        e_c = 2.45  # MPa

        with pytest.raises(ValueError):
            Form3Dot6CreepDeformationOfConcrete(phi_inf_t0=phi_inf_t0, sigma_c=sigma_c, e_c=e_c)

    def test_raise_error_when_negative_e_c_is_given(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        phi_inf_t0 = 0.34  # -
        sigma_c = 0.75  # MPa
        e_c = -2.45  # MPa

        with pytest.raises(ValueError):
            Form3Dot6CreepDeformationOfConcrete(phi_inf_t0=phi_inf_t0, sigma_c=sigma_c, e_c=e_c)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\epsilon_{cc}(\infty, t_0) = \varphi(\infty, t_0) \cdot ( \sigma_c / E_c ) = 0.340 \cdot ( 0.750 / 2.450 ) = 0.104",
            ),
            ("short", r"\epsilon_{cc}(\infty, t_0) = 0.104"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        phi_inf_t0 = 0.34  # -
        sigma_c = 0.75  # MPa
        e_c = 2.45  # MPa

        # Object to test
        form_3_6_latex = Form3Dot6CreepDeformationOfConcrete(phi_inf_t0=phi_inf_t0, sigma_c=sigma_c, e_c=e_c).latex()

        actual = {"complete": form_3_6_latex.complete, "short": form_3_6_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 3.13 from EN 1992-1-1:2004: Chapter 3 - Materials."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DAYS, DIMENSIONLESS


class Form3Dot13CoefficientTimeAutogeneShrinkage(Formula):
    """Class representing formula 3.13, which calculates the coefficient dependent on time for the autogene shrinkage."""

    source_document = EN_1992_1_1_2004
    label = "3.13"

    def __init__(
        self,
        t: DAYS,
    ) -> None:
        r"""[$\beta_{as}(t)$] Coefficient dependent on time in days for autogene shrinkage [$-$].

        EN 1992-1-1:2004 art.3.1.4(6) - Formula (3.13)

        Parameters
        ----------
        t : DAYS
            [$t$] Time in days [$days$].

        Returns
        -------
        None
        """
        super().__init__()
        self.t = t

    @staticmethod
    def _evaluate(
        t: DAYS,
    ) -> DIMENSIONLESS:
        r"""Evaluates the formula, for more information see the __init__ method."""
        if t < 0:
            raise ValueError(f"Invalid t: {t}. t cannot be negative")
        return 1 - np.exp(-0.2 * t**0.5)

    def latex(self, n: int = 2) -> LatexFormula:
        r"""Returns LatexFormula object for formula 3.13."""
        return LatexFormula(
            return_symbol=r"\beta_{as}(t)",
            result=f"{self:.{n}f}",
            equation=r"1 - \exp(-0.2 \cdot t^{0.5})",
            numeric_equation=rf"1 - \exp(-0.2 \cdot {self.t:.{n}f}^{{0.5}})",
            comparison_operator_label="=",
        )



"""Testing formula 3.27 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_27 import Form3Dot27IncreasedStrainLimitValue


class TestForm3Dot27IncreasedStrainLimitValue:
    """Validation for formula 3.27 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_ck = 12.2  # MPa
        sigma_2 = 0.08  # MPa
        epsilon_cu2 = 0.33  # -

        form_3_27 = Form3Dot27IncreasedStrainLimitValue(f_ck=f_ck, sigma_2=sigma_2, epsilon_cu2=epsilon_cu2)

        # Expected result, manually calculated
        manually_calculated_result = 0.331311

        assert form_3_27 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_ck_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        f_ck = -12.2  # MPa
        sigma_2 = 0.08  # MPa
        epsilon_cu2 = 0.33  # -

        with pytest.raises(ValueError):
            Form3Dot27IncreasedStrainLimitValue(f_ck=f_ck, sigma_2=sigma_2, epsilon_cu2=epsilon_cu2)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\epsilon_{cu2,c} = \epsilon_{cu2} + 0.2 \cdot \sigma_2 / f_{ck} = 0.330 + 0.2 \cdot 0.080 / 12.200 = 0.331",
            ),
            ("short", r"\epsilon_{cu2,c} = 0.331"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_ck = 12.2  # MPa
        sigma_2 = 0.08  # MPa
        epsilon_cu2 = 0.33  # -

        # Object to test
        form_3_27_latex = Form3Dot27IncreasedStrainLimitValue(f_ck=f_ck, sigma_2=sigma_2, epsilon_cu2=epsilon_cu2).latex()

        actual = {"complete": form_3_27_latex.complete, "short": form_3_27_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Tests for circular tube cross-section shapes."""

import numpy as np
import pytest

from blueprints.structural_sections.cross_section_tube import TubeCrossSection


class TestTubeCrossSection:
    """Tests for the TubeCrossSection class."""

    def test_area(self, tube_cross_section: TubeCrossSection) -> None:
        """Test the area property of the TubeCrossSection class."""
        expected_area = np.pi * (50.0**2 - 25.0**2)
        assert tube_cross_section.area == pytest.approx(expected=expected_area, rel=1e-3)

    def test_wall_thickness(self, tube_cross_section: TubeCrossSection) -> None:
        """Test the wall thickness property of the TubeCrossSection class."""
        expected_thickness = (100.0 - 50.0) / 2.0
        assert tube_cross_section.wall_thickness == pytest.approx(expected=expected_thickness, rel=1e-3)

    def test_polygon(self, tube_cross_section: TubeCrossSection) -> None:
        """Test the geometry property of the TubeCrossSection class."""
        polygon = tube_cross_section.polygon
        assert polygon.is_valid
        assert polygon.area == pytest.approx(expected=tube_cross_section.area, rel=1e-3)

    def test_invalid_outer_diameter(self) -> None:
        """Test initialization with an invalid outer diameter value."""
        with pytest.raises(ValueError, match="Outer diameter must be a positive value"):
            TubeCrossSection(name="InvalidOuter", outer_diameter=-10.0, inner_diameter=5.0, x=0.0, y=0.0)

    def test_invalid_inner_diameter(self) -> None:
        """Test initialization with an invalid inner diameter value."""
        with pytest.raises(ValueError, match="Inner diameter cannot be negative"):
            TubeCrossSection(name="InvalidInner", outer_diameter=10.0, inner_diameter=-5.0, x=0.0, y=0.0)

    def test_inner_diameter_greater_than_outer(self) -> None:
        """Test initialization with inner diameter greater than or equal to outer diameter."""
        with pytest.raises(ValueError, match="Inner diameter must be smaller than outer diameter"):
            TubeCrossSection(name="InvalidDiameters", outer_diameter=9.0, inner_diameter=10.0, x=0.0, y=0.0)

    def test_geometry(self, tube_cross_section: TubeCrossSection) -> None:
        """Test the geometry property of the TubeCrossSection class."""
        geometry = tube_cross_section.geometry()
        assert geometry is not None



"""Testing formula 6.48 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_48 import Form6Dot48NetAppliedPunchingForce
from blueprints.validations import NegativeValueError


class TestForm6Dot48NetAppliedPunchingForce:
    """Validation for formula 6.48 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 1000.0
        delta_v_ed = 200.0

        # Object to test
        formula = Form6Dot48NetAppliedPunchingForce(v_ed=v_ed, delta_v_ed=delta_v_ed)

        # Expected result, manually calculated
        manually_calculated_result = 800.0  # N

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("v_ed", "delta_v_ed"),
        [
            (-1000.0, 200.0),  # v_ed is negative
            (1000.0, -200.0),  # delta_v_ed is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, delta_v_ed: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot48NetAppliedPunchingForce(v_ed=v_ed, delta_v_ed=delta_v_ed)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{Ed,red} = V_{Ed} - \Delta V_{Ed} = 1000.000 - 200.000 = 800.000 \ N",
            ),
            ("short", r"V_{Ed,red} = 800.000 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 1000.0
        delta_v_ed = 200.0

        # Object to test
        latex = Form6Dot48NetAppliedPunchingForce(v_ed=v_ed, delta_v_ed=delta_v_ed).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."
